open Syntax;;

let verbose = ref 0;;

let atomh : (int,trm) Hashtbl.t = Hashtbl.create 100;;
let eigenvarforh : (string,trm) Hashtbl.t = Hashtbl.create 100;;

type sugg = SuggCut of trm | SuggInst of trm

let suggstack = ref [];;

let imp p q = Ap(Ap(Imp,p),q);;
let neg p = imp p False;;
let disj p q = imp (neg p) q;;
let conj p q = neg (imp p (neg q));;
let all b p = Ap(Forall(b),Lam(b,p));;
let ex b p = neg (all b (neg p));;
let eq b m n = Ap(Ap(Eq(b),m),n);;

  (*** Since this is a fake advisor, I am simply hard coding in some 'features' that tell me what problem it's looking at so I can suggest the right thing to do ***)
  (*** Hard Coded Example 1: (neq example) if x and y are different, there is a predicate p with p x and not p y. ***)
let recog_neq_example m =
  match m with
  | Ap(Forall(Ar(Base(v7),Prop)),Lam(Ar(Base(v6),Prop),Ap(Ap(Imp,Ap(DB(0,Ar(Base(v5),Prop)),Name(v3,Base(v4)))),Ap(DB(0,Ar(Base(v2),Prop)),Name(v0,Base(v1)))))) when v6 = v7 && v5 = v7 && v4 = v7 && v2 = v7 && v1 = v7  ->
      if !verbose > 9 then (Printf.printf "Suggesting inst \\x.x = %s\n" v3; flush stdout);
      suggstack := SuggInst(Lam(Base(v7),Ap(Ap(Eq(Base(v7)),DB(0,Base(v7))),Name(v3,Base(v4)))))::!suggstack
  | _ -> ();;

	(*** Hard Coded Example 2: manyeigens2 example ***)
let manyeigens2 = ref None;;
    
let recog_manyeigens2_example m =
  match m with
  | Ap(Ap(Imp,Ap(Forall(Base(v29)),Lam(Base(v28),Ap(Ap(Imp,Ap(Forall(Base(v27)),Lam(Base(v26),Ap(Ap(Imp,Ap(Ap(Ap(Name(v22,Ar(Base(v25),Ar(Base(v24),Ar(Base(v23),Prop)))),Ap(Ap(Name(v18,Ar(Base(v21),Ar(Base(v20),Base(v19)))),Ap(Ap(Name(v14,Ar(Base(v17),Ar(Base(v16),Base(v15)))),Name(v12,Base(v13))),Name(v10,Base(v11)))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Name(v4,Base(v5))),Name(v2,Base(v3))))),DB(1,Base(v1))),DB(0,Base(v0)))),False)))),False)))),False) when v29 = v28 && v29 = v27 && v29 = v26 && v29 = v24 && v29 = v23 && v25 = v21 && v25 = v20 && v25 = v19 && v25 = v17 && v25 = v16 && v25 = v15 && v18 = v14 && v25 = v13 && v25 = v11 && v12 = v10 && v25 = v9 && v25 = v8 && v25 = v7 && v18 = v6 && v25 = v5 && v12 = v4 && v25 = v3 && v12 = v2 && v29 = v1 && v29 = v0 ->
      if !verbose > 9 then (Printf.printf "manyeigens2 recognized with %s %s %s %s %s\n" v25 v29 v2 v14 v22; flush stdout);
      manyeigens2 := Some(v25,v29,v2,v14,v22) (** type a,type b, g:a, o:a>a>a, r:a>b>b>o **)
  | _ -> ();;

let manyeigens2_eigen_1_p x =
  match !manyeigens2 with
  | None -> false
  | Some(a,b,g,o,r) ->
      try
	let basetp_a_p y = y = a in
	let basetp_b_p y = y = b in
	let g_p y = y = g in
	let o_p y = y = o in
	let r_p y = y = r in
	match Hashtbl.find eigenvarforh x with
	| Ap(Forall(Base(v29)),Lam(Base(v28),Ap(Ap(Imp,Ap(Forall(Base(v27)),Lam(Base(v26),Ap(Ap(Imp,Ap(Ap(Ap(Name(v22,Ar(Base(v25),Ar(Base(v24),Ar(Base(v23),Prop)))),Ap(Ap(Name(v18,Ar(Base(v21),Ar(Base(v20),Base(v19)))),Ap(Ap(Name(v14,Ar(Base(v17),Ar(Base(v16),Base(v15)))),Name(v12,Base(v13))),Name(v10,Base(v11)))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Name(v4,Base(v5))),Name(v2,Base(v3))))),DB(1,Base(v1))),DB(0,Base(v0)))),False)))),False))) when basetp_b_p v29 && basetp_b_p v28 && basetp_b_p v27 && basetp_b_p v26 && basetp_a_p v25 && basetp_b_p v24 && basetp_b_p v23 && r_p v22 && basetp_a_p v21 && basetp_a_p v20 && basetp_a_p v19 && o_p v18 && basetp_a_p v17 && basetp_a_p v16 && basetp_a_p v15 && o_p v14 && basetp_a_p v13 && g_p v12 && basetp_a_p v11 && g_p v10 && basetp_a_p v9 && basetp_a_p v8 && basetp_a_p v7 && o_p v6 && basetp_a_p v5 && g_p v4 && basetp_a_p v3 && g_p v2 && basetp_b_p v1 && basetp_b_p v0 ->
	    true
	| _ -> false
      with _ -> false;;

let manyeigens2_eigen_2_p x =
  match !manyeigens2 with
  | None -> false
  | Some(a,b,g,o,r) ->
      try
	let basetp_a_p y = y = a in
	let basetp_b_p y = y = b in
	let g_p y = y = g in
	let r_p y = y = r in
	match Hashtbl.find eigenvarforh x with
	| Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Imp,Ap(Ap(Ap(Name(v5,Ar(Base(v8),Ar(Base(v7),Ar(Base(v6),Prop)))),Name(v3,Base(v4))),Name(v1,Base(v2))),DB(0,Base(v0)))),False))) when basetp_b_p v10 && basetp_b_p v9 && basetp_a_p v8 && basetp_b_p v7 && basetp_b_p v6 && r_p v5 && basetp_a_p v4 && g_p v3 && basetp_b_p v2 && manyeigens2_eigen_1_p v1 && basetp_b_p v0 ->
	    true
	| _ -> false
      with _ -> false;;

let manyeigens2_eigen_3_p x =
  match !manyeigens2 with
  | None -> false
  | Some(a,b,g,o,r) ->
      try
	let basetp_a_p y = y = a in
	let basetp_b_p y = y = b in
	let g_p y = y = g in
	let r_p y = y = r in
	match Hashtbl.find eigenvarforh x with
	| Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Imp,Ap(Ap(Ap(Name(v5,Ar(Base(v8),Ar(Base(v7),Ar(Base(v6),Prop)))),Name(v3,Base(v4))),Name(v1,Base(v2))),DB(0,Base(v0)))),False))) when basetp_b_p v10 && basetp_b_p v9 && basetp_a_p v8 && basetp_b_p v7 && basetp_b_p v6 && r_p v5 && basetp_a_p v4 && g_p v3 && basetp_b_p v2 && manyeigens2_eigen_2_p v1 && basetp_b_p v0 ->
	    true
	| _ -> false
      with _ -> false;;

let manyeigens2_eigen_4_p x =
  match !manyeigens2 with
  | None -> false
  | Some(a,b,g,o,r) ->
      try
	let basetp_a_p y = y = a in
	let basetp_b_p y = y = b in
	let g_p y = y = g in
	let r_p y = y = r in
	match Hashtbl.find eigenvarforh x with
	| Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Imp,Ap(Ap(Ap(Name(v5,Ar(Base(v8),Ar(Base(v7),Ar(Base(v6),Prop)))),Name(v3,Base(v4))),Name(v1,Base(v2))),DB(0,Base(v0)))),False))) when basetp_b_p v10 && basetp_b_p v9 && basetp_a_p v8 && basetp_b_p v7 && basetp_b_p v6 && r_p v5 && basetp_a_p v4 && g_p v3 && basetp_b_p v2 && manyeigens2_eigen_3_p v1 && basetp_b_p v0 ->
	    true
	| _ -> false
      with _ -> false;;

let manyeigens2_eigen_5_p x =
  match !manyeigens2 with
  | None -> false
  | Some(a,b,g,o,r) ->
      try
	let basetp_a_p y = y = a in
	let basetp_b_p y = y = b in
	let g_p y = y = g in
	let r_p y = y = r in
	match Hashtbl.find eigenvarforh x with
	| Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Imp,Ap(Ap(Ap(Name(v5,Ar(Base(v8),Ar(Base(v7),Ar(Base(v6),Prop)))),Name(v3,Base(v4))),Name(v1,Base(v2))),DB(0,Base(v0)))),False))) when basetp_b_p v10 && basetp_b_p v9 && basetp_a_p v8 && basetp_b_p v7 && basetp_b_p v6 && r_p v5 && basetp_a_p v4 && g_p v3 && basetp_b_p v2 && manyeigens2_eigen_4_p v1 && basetp_b_p v0 ->
	    true
	| _ -> false
      with _ -> false;;
  
  (*** Hard Coded Example 3: replacement implies separation example ***)
let processed_neg_sep_for_pred = ref None;;

let processed_repl = ref None;;

let magic_repl_sep_inst_suggested = ref false;;

let recog_repl_ish m =
  match m with
  | Ap(Forall(Base(_)),Lam(Base(_),Ap(Forall(Ar(Base(_),Ar(Base(_),Prop))),Lam(Ar(Base(_),Ar(Base(_),Prop)),Ap(Ap(Imp,Ap(Forall(Base(_)),Lam(Base(_),Ap(Ap(Imp,Ap(Ap(Name(mem,Ar(Base(a),Ar(Base(a2),Prop))),DB(0,Base(_))),DB(2,Base(_)))),Ap(Forall(Base(_)),Lam(Base(_),Ap(Forall(Base(_)),Lam(Base(_),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(DB(3,Ar(Base(_),Ar(Base(_),Prop))),DB(2,Base(_))),DB(1,Base(_)))),Ap(Ap(Imp,Ap(Ap(DB(3,Ar(Base(_),Ar(Base(_),Prop))),DB(2,Base(_))),DB(0,Base(_)))),False))),False)),Ap(Ap(Eq(Base(_)),DB(1,Base(_))),DB(0,Base(_)))))))))))),Ap(Ap(Imp,Ap(Forall(Base(_)),Lam(Base(_),Ap(Ap(Imp,Ap(Forall(Base(_)),Lam(Base(_),Ap(Ap(Eq(Prop),Ap(Ap(Name(mem2,Ar(Base(a3),Ar(Base(a4),Prop))),DB(0,Base(_))),DB(1,Base(_)))),Ap(Ap(Imp,Ap(Forall(Base(_)),Lam(Base(_),Ap(Ap(Imp,Ap(Ap(Name(_,Ar(Base(_),Ar(Base(_),Prop))),DB(0,Base(_))),DB(4,Base(_)))),Ap(Ap(Imp,Ap(Ap(DB(3,Ar(Base(_),Ar(Base(_),Prop))),DB(0,Base(_))),DB(1,Base(_)))),False))))),False))))),False)))),False)))))) when mem = mem2 && a = a2 && a = a3 && a = a4 ->
      processed_repl := Some(a,mem)
  | _ -> ();;

let recog_neg_sep_for_pred_ish m =
  match m with
  | Ap(Forall(Base(v17)),Lam(Base(v16),Ap(Ap(Imp,Ap(Forall(Base(v15)),Lam(Base(v14),Ap(Ap(Eq(Prop),Ap(Ap(Name(v11,Ar(Base(v13),Ar(Base(v12),Prop))),DB(0,Base(v10))),DB(1,Base(v9)))),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v6,Ar(Base(v8),Ar(Base(v7),Prop))),DB(0,Base(v5))),Name(v3,Base(v4)))),Ap(Ap(Imp,Ap(Name(v1,Ar(Base(v2),Prop)),DB(0,Base(v0)))),False))),False))))),False))) when v16 = v17 && v15 = v16 && v14 = v15 && v13 = v14 && v12 = v13 && v12 = v10 && v10 = v9 && v9 = v8 && v8 = v7 && v11 = v6 && v7 = v5 && v5 = v4 && v4 = v2 && v2 = v0 ->
      processed_neg_sep_for_pred := Some(v0,v6,v1,v3)
  | _ -> ();;
    
let repl_imp_sep_inst settp predname =
  Lam(settp,Lam(settp,Ap(Ap(And,Ap(Name(predname,Ar(settp,Prop)),DB(1,settp))),Ap(Ap(Eq(settp),DB(1,settp)),DB(0,settp)))));;

let suggest_repl_sep_inst () =
  if not !magic_repl_sep_inst_suggested then
    begin
      match !processed_neg_sep_for_pred with
      | None -> ()
      | Some(a,mem,pred,bdset) ->
	  match !processed_repl with
	  | None -> ()
	  | Some(a2,mem2) ->
	      if (a = a2) && (mem = mem2) then
		begin (*** suggest the bdset as an instantiation (of type a) and (\x y.p x & x = y) as an instantiation of type a>a>o ***)
		  suggstack := SuggInst(Name(bdset,Base(a)))::SuggInst(repl_imp_sep_inst (Base(a)) pred)::!suggstack;
		  if !verbose > 1 then (Printf.printf "added repl imp sep inst to suggstack (a = %s, mem = %s, pred = %s, bdset = %s)\n" a mem pred bdset; flush stdout);
		  magic_repl_sep_inst_suggested := true
		end
    end;;

let adapted_process_prop_1_repl_imp_sep a mem pred bdset p m =
  let basetp_i_p v = v = a in
  let p_p v = v = pred in
  let mem_p v = v = mem in
  let a_p v = v = bdset in
  let eigen__0_p v = a_p v in
  let eigen__1_p v = p_p v in
  let eigen__2_p v =
    try
      let m = Hashtbl.find eigenvarforh v in
      match m with
      | Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Imp,Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Eq(Prop),Ap(Ap(Name(v16,Ar(Base(v18),Ar(Base(v17),Prop))),DB(0,Base(v15))),DB(1,Base(v14)))),Ap(Ap(Imp,Ap(Forall(Base(v13)),Lam(Base(v12),Ap(Ap(Imp,Ap(Ap(Name(v9,Ar(Base(v11),Ar(Base(v10),Prop))),DB(0,Base(v8))),Name(v6,Base(v7)))),Ap(Ap(Imp,Ap(Name(v4,Ar(Base(v5),Prop)),DB(0,Base(v3)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v2)),DB(0,Base(v1))),DB(1,Base(v0)))),False)))))),False))))),False))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && mem_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && mem_p v9 && basetp_i_p v8 && basetp_i_p v7 && eigen__0_p v6 && basetp_i_p v5 && eigen__1_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 ->
	  true
      | _ ->
	  false
    with Not_found -> false in
  let eigen__3_p v = try match Hashtbl.find eigenvarforh v with Ap(Forall(Base(v26)),Lam(Base(v25),Ap(Ap(Imp,Ap(Ap(Name(v22,Ar(Base(v24),Ar(Base(v23),Prop))),DB(0,Base(v21))),Name(v19,Base(v20)))),Ap(Forall(Base(v18)),Lam(Base(v17),Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v13,Ar(Base(v14),Prop)),DB(2,Base(v12)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v11)),DB(2,Base(v10))),DB(1,Base(v9)))),False))),False)),Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),DB(2,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v5)),DB(2,Base(v4))),DB(0,Base(v3)))),False)))),False)),Ap(Ap(Eq(Base(v2)),DB(1,Base(v1))),DB(0,Base(v0))))))))))) when basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && mem_p v22 && basetp_i_p v21 && basetp_i_p v20 && eigen__0_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && eigen__1_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> true | _ -> false with Not_found -> false in
  let eigen__4_p v = try match Hashtbl.find eigenvarforh v with Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v17,Ar(Base(v18),Prop)),Name(v15,Base(v16)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v14)),Name(v12,Base(v13))),DB(1,Base(v11)))),False))),False)),Ap(Ap(Imp,Ap(Name(v9,Ar(Base(v10),Prop)),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v6)),Name(v4,Base(v5))),DB(0,Base(v3)))),False)))),False)),Ap(Ap(Eq(Base(v2)),DB(1,Base(v1))),DB(0,Base(v0)))))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && eigen__1_p v17 && basetp_i_p v16 && eigen__3_p v15 && basetp_i_p v14 && basetp_i_p v13 && eigen__3_p v12 && basetp_i_p v11 && basetp_i_p v10 && eigen__1_p v9 && basetp_i_p v8 && eigen__3_p v7 && basetp_i_p v6 && basetp_i_p v5 && eigen__3_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> true | _ -> false with Not_found -> false in
  let eigen__5_p v = try match Hashtbl.find eigenvarforh v with Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v19,Ar(Base(v20),Prop)),Name(v17,Base(v18)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v16)),Name(v14,Base(v15))),Name(v12,Base(v13)))),False))),False)),Ap(Ap(Imp,Ap(Name(v10,Ar(Base(v11),Prop)),Name(v8,Base(v9)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),Name(v5,Base(v6))),DB(0,Base(v4)))),False)))),False)),Ap(Ap(Eq(Base(v3)),Name(v1,Base(v2))),DB(0,Base(v0)))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && eigen__1_p v19 && basetp_i_p v18 && eigen__3_p v17 && basetp_i_p v16 && basetp_i_p v15 && eigen__3_p v14 && basetp_i_p v13 && eigen__4_p v12 && basetp_i_p v11 && eigen__1_p v10 && basetp_i_p v9 && eigen__3_p v8 && basetp_i_p v7 && basetp_i_p v6 && eigen__3_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && eigen__4_p v1 && basetp_i_p v0 -> true | _ -> false with Not_found -> false in
  let eigen__6_p v = try match Hashtbl.find eigenvarforh v with Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Eq(Prop),Ap(Ap(Name(v12,Ar(Base(v14),Ar(Base(v13),Prop))),DB(0,Base(v11))),Name(v9,Base(v10)))),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v6,Ar(Base(v8),Ar(Base(v7),Prop))),DB(0,Base(v5))),Name(v3,Base(v4)))),Ap(Ap(Imp,Ap(Name(v1,Ar(Base(v2),Prop)),DB(0,Base(v0)))),False))),False)))) when basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && mem_p v12 && basetp_i_p v11 && basetp_i_p v10 && eigen__2_p v9 && basetp_i_p v8 && basetp_i_p v7 && mem_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && eigen__1_p v1 && basetp_i_p v0 -> true | _ -> false with Not_found -> false in
  let eigen__7_p v = try match Hashtbl.find eigenvarforh v with Ap(Forall(Base(v14)),Lam(Base(v13),Ap(Ap(Imp,Ap(Ap(Name(v10,Ar(Base(v12),Ar(Base(v11),Prop))),DB(0,Base(v9))),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Name(v5,Ar(Base(v6),Prop)),DB(0,Base(v4)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v3)),DB(0,Base(v2))),Name(v0,Base(v1)))),False))))) when basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && mem_p v10 && basetp_i_p v9 && basetp_i_p v8 && eigen__0_p v7 && basetp_i_p v6 && eigen__1_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__6_p v0 -> true | _ -> false with Not_found -> false in
  match m with
  | Ap(Forall(Ar(Base(v47),Ar(Base(v46),Prop))),Lam(Ar(Base(v45),Ar(Base(v44),Prop)),Ap(Ap(Imp,Ap(Forall(Base(v43)),Lam(Base(v42),Ap(Ap(Imp,Ap(Ap(Name(v39,Ar(Base(v41),Ar(Base(v40),Prop))),DB(0,Base(v38))),Name(v36,Base(v37)))),Ap(Forall(Base(v35)),Lam(Base(v34),Ap(Forall(Base(v33)),Lam(Base(v32),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(DB(3,Ar(Base(v31),Ar(Base(v30),Prop))),DB(2,Base(v29))),DB(1,Base(v28)))),Ap(Ap(Imp,Ap(Ap(DB(3,Ar(Base(v27),Ar(Base(v26),Prop))),DB(2,Base(v25))),DB(0,Base(v24)))),False))),False)),Ap(Ap(Eq(Base(v23)),DB(1,Base(v22))),DB(0,Base(v21)))))))))))),Ap(Ap(Imp,Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Imp,Ap(Forall(Base(v18)),Lam(Base(v17),Ap(Ap(Eq(Prop),Ap(Ap(Name(v14,Ar(Base(v16),Ar(Base(v15),Prop))),DB(0,Base(v13))),DB(1,Base(v12)))),Ap(Ap(Imp,Ap(Forall(Base(v11)),Lam(Base(v10),Ap(Ap(Imp,Ap(Ap(Name(v7,Ar(Base(v9),Ar(Base(v8),Prop))),DB(0,Base(v6))),Name(v4,Base(v5)))),Ap(Ap(Imp,Ap(Ap(DB(3,Ar(Base(v3),Ar(Base(v2),Prop))),DB(0,Base(v1))),DB(1,Base(v0)))),False))))),False))))),False)))),False)))) when basetp_i_p v47 && basetp_i_p v46 && basetp_i_p v45 && basetp_i_p v44 && basetp_i_p v43 && basetp_i_p v42 && basetp_i_p v41 && basetp_i_p v40 && mem_p v39 && basetp_i_p v38 && basetp_i_p v37 && eigen__0_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && basetp_i_p v30 && basetp_i_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && mem_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && mem_p v7 && basetp_i_p v6 && basetp_i_p v5 && eigen__0_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v49)),Lam(Base(v48),Ap(Ap(Imp,Ap(Ap(Name(v45,Ar(Base(v47),Ar(Base(v46),Prop))),DB(0,Base(v44))),Name(v42,Base(v43)))),Ap(Forall(Base(v41)),Lam(Base(v40),Ap(Forall(Base(v39)),Lam(Base(v38),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v36,Ar(Base(v37),Prop)),DB(2,Base(v35)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v34)),DB(2,Base(v33))),DB(1,Base(v32)))),False))),False)),Ap(Ap(Imp,Ap(Name(v30,Ar(Base(v31),Prop)),DB(2,Base(v29)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v28)),DB(2,Base(v27))),DB(0,Base(v26)))),False)))),False)),Ap(Ap(Eq(Base(v25)),DB(1,Base(v24))),DB(0,Base(v23)))))))))))),Ap(Ap(Imp,Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Imp,Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Eq(Prop),Ap(Ap(Name(v16,Ar(Base(v18),Ar(Base(v17),Prop))),DB(0,Base(v15))),DB(1,Base(v14)))),Ap(Ap(Imp,Ap(Forall(Base(v13)),Lam(Base(v12),Ap(Ap(Imp,Ap(Ap(Name(v9,Ar(Base(v11),Ar(Base(v10),Prop))),DB(0,Base(v8))),Name(v6,Base(v7)))),Ap(Ap(Imp,Ap(Name(v4,Ar(Base(v5),Prop)),DB(0,Base(v3)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v2)),DB(0,Base(v1))),DB(1,Base(v0)))),False)))))),False))))),False)))),False)) when basetp_i_p v49 && basetp_i_p v48 && basetp_i_p v47 && basetp_i_p v46 && mem_p v45 && basetp_i_p v44 && basetp_i_p v43 && eigen__0_p v42 && basetp_i_p v41 && basetp_i_p v40 && basetp_i_p v39 && basetp_i_p v38 && basetp_i_p v37 && eigen__1_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && eigen__1_p v30 && basetp_i_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && mem_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && mem_p v9 && basetp_i_p v8 && basetp_i_p v7 && eigen__0_p v6 && basetp_i_p v5 && eigen__1_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v26)),Lam(Base(v25),Ap(Ap(Imp,Ap(Ap(Name(v22,Ar(Base(v24),Ar(Base(v23),Prop))),DB(0,Base(v21))),Name(v19,Base(v20)))),Ap(Forall(Base(v18)),Lam(Base(v17),Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v13,Ar(Base(v14),Prop)),DB(2,Base(v12)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v11)),DB(2,Base(v10))),DB(1,Base(v9)))),False))),False)),Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),DB(2,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v5)),DB(2,Base(v4))),DB(0,Base(v3)))),False)))),False)),Ap(Ap(Eq(Base(v2)),DB(1,Base(v1))),DB(0,Base(v0)))))))))))),False) when basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && mem_p v22 && basetp_i_p v21 && basetp_i_p v20 && eigen__0_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && eigen__1_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Imp,Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Eq(Prop),Ap(Ap(Name(v16,Ar(Base(v18),Ar(Base(v17),Prop))),DB(0,Base(v15))),DB(1,Base(v14)))),Ap(Ap(Imp,Ap(Forall(Base(v13)),Lam(Base(v12),Ap(Ap(Imp,Ap(Ap(Name(v9,Ar(Base(v11),Ar(Base(v10),Prop))),DB(0,Base(v8))),Name(v6,Base(v7)))),Ap(Ap(Imp,Ap(Name(v4,Ar(Base(v5),Prop)),DB(0,Base(v3)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v2)),DB(0,Base(v1))),DB(1,Base(v0)))),False)))))),False))))),False)))),False) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && mem_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && mem_p v9 && basetp_i_p v8 && basetp_i_p v7 && eigen__0_p v6 && basetp_i_p v5 && eigen__1_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Forall(Base(v21)),Lam(Base(v20),Ap(Ap(Eq(Prop),Ap(Ap(Name(v17,Ar(Base(v19),Ar(Base(v18),Prop))),DB(0,Base(v16))),Name(v14,Base(v15)))),Ap(Ap(Imp,Ap(Forall(Base(v13)),Lam(Base(v12),Ap(Ap(Imp,Ap(Ap(Name(v9,Ar(Base(v11),Ar(Base(v10),Prop))),DB(0,Base(v8))),Name(v6,Base(v7)))),Ap(Ap(Imp,Ap(Name(v4,Ar(Base(v5),Prop)),DB(0,Base(v3)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v2)),DB(0,Base(v1))),DB(1,Base(v0)))),False)))))),False)))) when basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && mem_p v17 && basetp_i_p v16 && basetp_i_p v15 && eigen__2_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && mem_p v9 && basetp_i_p v8 && basetp_i_p v7 && eigen__0_p v6 && basetp_i_p v5 && eigen__1_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 ->
      suggstack := SuggInst(Name(v14,Base(v15)))::!suggstack; (*** this is the set whose existence replacement ensures, which is the one we need to use to prove separation holds -- suggest the instantiation to satallax to prevent it from needing to find out it should consider it ***)
      -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v27,Ar(Base(v29),Ar(Base(v28),Prop))),Name(v25,Base(v26))),Name(v23,Base(v24)))),Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v17,Ar(Base(v18),Prop)),Name(v15,Base(v16)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v14)),Name(v12,Base(v13))),DB(1,Base(v11)))),False))),False)),Ap(Ap(Imp,Ap(Name(v9,Ar(Base(v10),Prop)),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v6)),Name(v4,Base(v5))),DB(0,Base(v3)))),False)))),False)),Ap(Ap(Eq(Base(v2)),DB(1,Base(v1))),DB(0,Base(v0)))))))))),False)),False) when basetp_i_p v29 && basetp_i_p v28 && mem_p v27 && basetp_i_p v26 && eigen__3_p v25 && basetp_i_p v24 && eigen__0_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && eigen__1_p v17 && basetp_i_p v16 && eigen__3_p v15 && basetp_i_p v14 && basetp_i_p v13 && eigen__3_p v12 && basetp_i_p v11 && basetp_i_p v10 && eigen__1_p v9 && basetp_i_p v8 && eigen__3_p v7 && basetp_i_p v6 && basetp_i_p v5 && eigen__3_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Forall(Base(v17)),Lam(Base(v16),Ap(Ap(Imp,Ap(Forall(Base(v15)),Lam(Base(v14),Ap(Ap(Eq(Prop),Ap(Ap(Name(v11,Ar(Base(v13),Ar(Base(v12),Prop))),DB(0,Base(v10))),DB(1,Base(v9)))),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v6,Ar(Base(v8),Ar(Base(v7),Prop))),DB(0,Base(v5))),Name(v3,Base(v4)))),Ap(Ap(Imp,Ap(Name(v1,Ar(Base(v2),Prop)),DB(0,Base(v0)))),False))),False))))),False))) when basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && mem_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && mem_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && eigen__1_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v27,Ar(Base(v29),Ar(Base(v28),Prop))),Name(v25,Base(v26))),Name(v23,Base(v24)))),Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v17,Ar(Base(v18),Prop)),Name(v15,Base(v16)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v14)),Name(v12,Base(v13))),DB(1,Base(v11)))),False))),False)),Ap(Ap(Imp,Ap(Name(v9,Ar(Base(v10),Prop)),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v6)),Name(v4,Base(v5))),DB(0,Base(v3)))),False)))),False)),Ap(Ap(Eq(Base(v2)),DB(1,Base(v1))),DB(0,Base(v0)))))))))),False) when basetp_i_p v29 && basetp_i_p v28 && mem_p v27 && basetp_i_p v26 && eigen__3_p v25 && basetp_i_p v24 && eigen__0_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && eigen__1_p v17 && basetp_i_p v16 && eigen__3_p v15 && basetp_i_p v14 && basetp_i_p v13 && eigen__3_p v12 && basetp_i_p v11 && basetp_i_p v10 && eigen__1_p v9 && basetp_i_p v8 && eigen__3_p v7 && basetp_i_p v6 && basetp_i_p v5 && eigen__3_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Name(v4,Ar(Base(v6),Ar(Base(v5),Prop))),Name(v2,Base(v3))),Name(v0,Base(v1))) when basetp_i_p v6 && basetp_i_p v5 && mem_p v4 && basetp_i_p v3 && eigen__3_p v2 && basetp_i_p v1 && eigen__0_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Forall(Base(v20)),Lam(Base(v19),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v17,Ar(Base(v18),Prop)),Name(v15,Base(v16)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v14)),Name(v12,Base(v13))),DB(1,Base(v11)))),False))),False)),Ap(Ap(Imp,Ap(Name(v9,Ar(Base(v10),Prop)),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v6)),Name(v4,Base(v5))),DB(0,Base(v3)))),False)))),False)),Ap(Ap(Eq(Base(v2)),DB(1,Base(v1))),DB(0,Base(v0))))))))),False) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && eigen__1_p v17 && basetp_i_p v16 && eigen__3_p v15 && basetp_i_p v14 && basetp_i_p v13 && eigen__3_p v12 && basetp_i_p v11 && basetp_i_p v10 && eigen__1_p v9 && basetp_i_p v8 && eigen__3_p v7 && basetp_i_p v6 && basetp_i_p v5 && eigen__3_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v19,Ar(Base(v20),Prop)),Name(v17,Base(v18)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v16)),Name(v14,Base(v15))),Name(v12,Base(v13)))),False))),False)),Ap(Ap(Imp,Ap(Name(v10,Ar(Base(v11),Prop)),Name(v8,Base(v9)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),Name(v5,Base(v6))),DB(0,Base(v4)))),False)))),False)),Ap(Ap(Eq(Base(v3)),Name(v1,Base(v2))),DB(0,Base(v0))))))),False) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && eigen__1_p v19 && basetp_i_p v18 && eigen__3_p v17 && basetp_i_p v16 && basetp_i_p v15 && eigen__3_p v14 && basetp_i_p v13 && eigen__4_p v12 && basetp_i_p v11 && eigen__1_p v10 && basetp_i_p v9 && eigen__3_p v8 && basetp_i_p v7 && basetp_i_p v6 && eigen__3_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && eigen__4_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v21,Ar(Base(v22),Prop)),Name(v19,Base(v20)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v18)),Name(v16,Base(v17))),Name(v14,Base(v15)))),False))),False)),Ap(Ap(Imp,Ap(Name(v12,Ar(Base(v13),Prop)),Name(v10,Base(v11)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v9)),Name(v7,Base(v8))),Name(v5,Base(v6)))),False)))),False)),Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1))))),False) when basetp_i_p v22 && eigen__1_p v21 && basetp_i_p v20 && eigen__3_p v19 && basetp_i_p v18 && basetp_i_p v17 && eigen__3_p v16 && basetp_i_p v15 && eigen__4_p v14 && basetp_i_p v13 && eigen__1_p v12 && basetp_i_p v11 && eigen__3_p v10 && basetp_i_p v9 && basetp_i_p v8 && eigen__3_p v7 && basetp_i_p v6 && eigen__5_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__4_p v2 && basetp_i_p v1 && eigen__5_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v16,Ar(Base(v17),Prop)),Name(v14,Base(v15)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v13)),Name(v11,Base(v12))),Name(v9,Base(v10)))),False))),False)),Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False)))),False) when basetp_i_p v17 && eigen__1_p v16 && basetp_i_p v15 && eigen__3_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__3_p v11 && basetp_i_p v10 && eigen__4_p v9 && basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__3_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__3_p v2 && basetp_i_p v1 && eigen__5_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False) when basetp_i_p v4 && basetp_i_p v3 && eigen__4_p v2 && basetp_i_p v1 && eigen__5_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v16,Ar(Base(v17),Prop)),Name(v14,Base(v15)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v13)),Name(v11,Base(v12))),Name(v9,Base(v10)))),False))),False) when basetp_i_p v17 && eigen__1_p v16 && basetp_i_p v15 && eigen__3_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__3_p v11 && basetp_i_p v10 && eigen__4_p v9 -> -1
  | Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False)) when basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__3_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__3_p v2 && basetp_i_p v1 && eigen__5_p v0 -> -1
  | Ap(Name(v16,Ar(Base(v17),Prop)),Name(v14,Base(v15))) when basetp_i_p v17 && eigen__1_p v16 && basetp_i_p v15 && eigen__3_p v14 -> -1
  | Ap(Ap(Eq(Base(v13)),Name(v11,Base(v12))),Name(v9,Base(v10))) when basetp_i_p v13 && basetp_i_p v12 && eigen__3_p v11 && basetp_i_p v10 && eigen__4_p v9 -> -1
  | Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),False) when basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__3_p v5 -> -1
  | Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1))) when basetp_i_p v4 && basetp_i_p v3 && eigen__3_p v2 && basetp_i_p v1 && eigen__5_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Eq(Prop),Ap(Ap(Name(v12,Ar(Base(v14),Ar(Base(v13),Prop))),DB(0,Base(v11))),Name(v9,Base(v10)))),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v6,Ar(Base(v8),Ar(Base(v7),Prop))),DB(0,Base(v5))),Name(v3,Base(v4)))),Ap(Ap(Imp,Ap(Name(v1,Ar(Base(v2),Prop)),DB(0,Base(v0)))),False))),False))))),False) when basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && mem_p v12 && basetp_i_p v11 && basetp_i_p v10 && eigen__2_p v9 && basetp_i_p v8 && basetp_i_p v7 && mem_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && eigen__1_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Prop),Ap(Ap(Name(v15,Ar(Base(v17),Ar(Base(v16),Prop))),Name(v13,Base(v14))),Name(v11,Base(v12)))),Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v8,Ar(Base(v10),Ar(Base(v9),Prop))),Name(v6,Base(v7))),Name(v4,Base(v5)))),Ap(Ap(Imp,Ap(Name(v2,Ar(Base(v3),Prop)),Name(v0,Base(v1)))),False))),False))),False) when basetp_i_p v17 && basetp_i_p v16 && mem_p v15 && basetp_i_p v14 && eigen__6_p v13 && basetp_i_p v12 && eigen__2_p v11 && basetp_i_p v10 && basetp_i_p v9 && mem_p v8 && basetp_i_p v7 && eigen__6_p v6 && basetp_i_p v5 && eigen__0_p v4 && basetp_i_p v3 && eigen__1_p v2 && basetp_i_p v1 && eigen__6_p v0 ->
      suggstack := SuggInst(Name(v13,Base(v14)))::!suggstack; (*** this is the element being used to test the separation property, which we need to instantiate into the property we of the set got from replacement ***)
      -1
  | Ap(Ap(Name(v15,Ar(Base(v17),Ar(Base(v16),Prop))),Name(v13,Base(v14))),Name(v11,Base(v12))) when basetp_i_p v17 && basetp_i_p v16 && mem_p v15 && basetp_i_p v14 && eigen__6_p v13 && basetp_i_p v12 && eigen__2_p v11 -> -1
  | Ap(Ap(Imp,Ap(Ap(Name(v15,Ar(Base(v17),Ar(Base(v16),Prop))),Name(v13,Base(v14))),Name(v11,Base(v12)))),False) when basetp_i_p v17 && basetp_i_p v16 && mem_p v15 && basetp_i_p v14 && eigen__6_p v13 && basetp_i_p v12 && eigen__2_p v11 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v8,Ar(Base(v10),Ar(Base(v9),Prop))),Name(v6,Base(v7))),Name(v4,Base(v5)))),Ap(Ap(Imp,Ap(Name(v2,Ar(Base(v3),Prop)),Name(v0,Base(v1)))),False))),False) when basetp_i_p v10 && basetp_i_p v9 && mem_p v8 && basetp_i_p v7 && eigen__6_p v6 && basetp_i_p v5 && eigen__0_p v4 && basetp_i_p v3 && eigen__1_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Name(v8,Ar(Base(v10),Ar(Base(v9),Prop))),Name(v6,Base(v7))),Name(v4,Base(v5)))),Ap(Ap(Imp,Ap(Name(v2,Ar(Base(v3),Prop)),Name(v0,Base(v1)))),False)) when basetp_i_p v10 && basetp_i_p v9 && mem_p v8 && basetp_i_p v7 && eigen__6_p v6 && basetp_i_p v5 && eigen__0_p v4 && basetp_i_p v3 && eigen__1_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Name(v8,Ar(Base(v10),Ar(Base(v9),Prop))),Name(v6,Base(v7))),Name(v4,Base(v5))) when basetp_i_p v10 && basetp_i_p v9 && mem_p v8 && basetp_i_p v7 && eigen__6_p v6 && basetp_i_p v5 && eigen__0_p v4 -> -1
  | Ap(Ap(Imp,Ap(Ap(Name(v8,Ar(Base(v10),Ar(Base(v9),Prop))),Name(v6,Base(v7))),Name(v4,Base(v5)))),False) when basetp_i_p v10 && basetp_i_p v9 && mem_p v8 && basetp_i_p v7 && eigen__6_p v6 && basetp_i_p v5 && eigen__0_p v4 -> -1
  | Ap(Name(v2,Ar(Base(v3),Prop)),Name(v0,Base(v1))) when basetp_i_p v3 && eigen__1_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Imp,Ap(Name(v2,Ar(Base(v3),Prop)),Name(v0,Base(v1)))),False) when basetp_i_p v3 && eigen__1_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Eq(Prop),Ap(Ap(Name(v19,Ar(Base(v21),Ar(Base(v20),Prop))),Name(v17,Base(v18))),Name(v15,Base(v16)))),Ap(Ap(Imp,Ap(Forall(Base(v14)),Lam(Base(v13),Ap(Ap(Imp,Ap(Ap(Name(v10,Ar(Base(v12),Ar(Base(v11),Prop))),DB(0,Base(v9))),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Name(v5,Ar(Base(v6),Prop)),DB(0,Base(v4)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v3)),DB(0,Base(v2))),Name(v0,Base(v1)))),False)))))),False)) when basetp_i_p v21 && basetp_i_p v20 && mem_p v19 && basetp_i_p v18 && eigen__6_p v17 && basetp_i_p v16 && eigen__2_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && mem_p v10 && basetp_i_p v9 && basetp_i_p v8 && eigen__0_p v7 && basetp_i_p v6 && eigen__1_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Name(v19,Ar(Base(v21),Ar(Base(v20),Prop))),Name(v17,Base(v18))),Name(v15,Base(v16))) when basetp_i_p v21 && basetp_i_p v20 && mem_p v19 && basetp_i_p v18 && eigen__6_p v17 && basetp_i_p v16 && eigen__2_p v15 -> -1
  | Ap(Ap(Imp,Ap(Ap(Name(v19,Ar(Base(v21),Ar(Base(v20),Prop))),Name(v17,Base(v18))),Name(v15,Base(v16)))),False) when basetp_i_p v21 && basetp_i_p v20 && mem_p v19 && basetp_i_p v18 && eigen__6_p v17 && basetp_i_p v16 && eigen__2_p v15 -> -1
  | Ap(Forall(Base(v14)),Lam(Base(v13),Ap(Ap(Imp,Ap(Ap(Name(v10,Ar(Base(v12),Ar(Base(v11),Prop))),DB(0,Base(v9))),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Name(v5,Ar(Base(v6),Prop)),DB(0,Base(v4)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v3)),DB(0,Base(v2))),Name(v0,Base(v1)))),False))))) when basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && mem_p v10 && basetp_i_p v9 && basetp_i_p v8 && eigen__0_p v7 && basetp_i_p v6 && eigen__1_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v14)),Lam(Base(v13),Ap(Ap(Imp,Ap(Ap(Name(v10,Ar(Base(v12),Ar(Base(v11),Prop))),DB(0,Base(v9))),Name(v7,Base(v8)))),Ap(Ap(Imp,Ap(Name(v5,Ar(Base(v6),Prop)),DB(0,Base(v4)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v3)),DB(0,Base(v2))),Name(v0,Base(v1)))),False)))))),False) when basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && mem_p v10 && basetp_i_p v9 && basetp_i_p v8 && eigen__0_p v7 && basetp_i_p v6 && eigen__1_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Name(v13,Ar(Base(v15),Ar(Base(v14),Prop))),Name(v11,Base(v12))),Name(v9,Base(v10)))),Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False))) when basetp_i_p v15 && basetp_i_p v14 && mem_p v13 && basetp_i_p v12 && eigen__6_p v11 && basetp_i_p v10 && eigen__0_p v9 && basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__6_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__6_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Name(v13,Ar(Base(v15),Ar(Base(v14),Prop))),Name(v11,Base(v12))),Name(v9,Base(v10))) when basetp_i_p v15 && basetp_i_p v14 && mem_p v13 && basetp_i_p v12 && eigen__6_p v11 && basetp_i_p v10 -> -1
  | Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False)) when basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__6_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__6_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),False) when basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__6_p v5 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False) when basetp_i_p v4 && basetp_i_p v3 && eigen__6_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Name(v13,Ar(Base(v15),Ar(Base(v14),Prop))),Name(v11,Base(v12))),Name(v9,Base(v10)))),Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False)))),False) when basetp_i_p v15 && basetp_i_p v14 && mem_p v13 && basetp_i_p v12 && eigen__7_p v11 && basetp_i_p v10 && eigen__0_p v9 && basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__7_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__7_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Ap(Name(v13,Ar(Base(v15),Ar(Base(v14),Prop))),Name(v11,Base(v12))),Name(v9,Base(v10))) when basetp_i_p v15 && basetp_i_p v14 && mem_p v13 && basetp_i_p v12 && eigen__7_p v11 && basetp_i_p v10 && eigen__0_p v9 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6)))),Ap(Ap(Imp,Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1)))),False))),False) when basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__7_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__7_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | Ap(Name(v7,Ar(Base(v8),Prop)),Name(v5,Base(v6))) when basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && eigen__7_p v5 -> -1
  | Ap(Ap(Eq(Base(v4)),Name(v2,Base(v3))),Name(v0,Base(v1))) when basetp_i_p v4 && basetp_i_p v3 && eigen__7_p v2 && basetp_i_p v1 && eigen__6_p v0 -> -1
  | _ ->
      if !verbose > 9 then (Printf.printf "Looks bad %s\n%s\n" (trm_str m) (trm_ocaml_match m); flush stdout);
      p + 500;;

  (*** Hard Coded Example 4: injcantor example ***)
let injcantor = ref None;;
    
let recog_injcantor m =
  match m with
  | Ap(Ap(Imp,Ap(Forall(Ar(Ar(Base(v17),Prop),Base(v16))),Lam(Ar(Ar(Base(v15),Prop),Base(v14)),Ap(Ap(Imp,Ap(Forall(Ar(Base(v13),Prop)),Lam(Ar(Base(v12),Prop),Ap(Forall(Ar(Base(v11),Prop)),Lam(Ar(Base(v10),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v9)),Ap(DB(2,Ar(Ar(Base(v8),Prop),Base(v7))),DB(1,Ar(Base(v6),Prop)))),Ap(DB(2,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v3),Prop))))),Ap(Ap(Eq(Ar(Base(v2),Prop)),DB(1,Ar(Base(v1),Prop))),DB(0,Ar(Base(v0),Prop))))))))),False)))),False) when v17 = v16 && v17 = v15 && v17 = v14 && v17 = v13 && v17 = v12 && v17 = v11 && v17 = v10 && v17 = v9 && v17 = v8 && v17 = v7 && v17 = v6 && v17 = v5 && v17 = v4 && v17 = v3 && v17 = v2 && v17 = v1 && v17 = v0 ->
      injcantor := Some(v17)
  | _ -> ();;

let adapted_process_prop_1_injcantor a p m =
  let basetp_i_p v = v = a in
  let eigen__0_p v =
    try
      let m = Hashtbl.find eigenvarforh v in
      match m with
      | Ap(Forall(Ar(Ar(Base(v17),Prop),Base(v16))),Lam(Ar(Ar(Base(v15),Prop),Base(v14)),Ap(Ap(Imp,Ap(Forall(Ar(Base(v13),Prop)),Lam(Ar(Base(v12),Prop),Ap(Forall(Ar(Base(v11),Prop)),Lam(Ar(Base(v10),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v9)),Ap(DB(2,Ar(Ar(Base(v8),Prop),Base(v7))),DB(1,Ar(Base(v6),Prop)))),Ap(DB(2,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v3),Prop))))),Ap(Ap(Eq(Ar(Base(v2),Prop)),DB(1,Ar(Base(v1),Prop))),DB(0,Ar(Base(v0),Prop))))))))),False))) when basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  let eigen__1_p v =
    try
      let m = Hashtbl.find eigenvarforh v in
      match m with
      | Ap(Forall(Ar(Base(v35),Prop)),Lam(Ar(Base(v34),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v33)),Ap(Name(v30,Ar(Ar(Base(v32),Prop),Base(v31))),Lam(Base(v29),Ap(Ap(Imp,Ap(Forall(Ar(Base(v28),Prop)),Lam(Ar(Base(v27),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v26)),DB(1,Base(v25))),Ap(Name(v22,Ar(Ar(Base(v24),Prop),Base(v23))),DB(0,Ar(Base(v21),Prop))))),Ap(DB(0,Ar(Base(v20),Prop)),DB(1,Base(v19))))))),False)))),Ap(Name(v16,Ar(Ar(Base(v18),Prop),Base(v17))),DB(0,Ar(Base(v15),Prop))))),Ap(DB(0,Ar(Base(v14),Prop)),Ap(Name(v11,Ar(Ar(Base(v13),Prop),Base(v12))),Lam(Base(v10),Ap(Ap(Imp,Ap(Forall(Ar(Base(v9),Prop)),Lam(Ar(Base(v8),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),DB(1,Base(v6))),Ap(Name(v3,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v2),Prop))))),Ap(DB(0,Ar(Base(v1),Prop)),DB(1,Base(v0))))))),False))))))) when basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && eigen__0_p v30 && basetp_i_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && eigen__0_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && eigen__0_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__0_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  match m with
  | Ap(Ap(Imp,Ap(Forall(Ar(Ar(Base(v17),Prop),Base(v16))),Lam(Ar(Ar(Base(v15),Prop),Base(v14)),Ap(Ap(Imp,Ap(Forall(Ar(Base(v13),Prop)),Lam(Ar(Base(v12),Prop),Ap(Forall(Ar(Base(v11),Prop)),Lam(Ar(Base(v10),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v9)),Ap(DB(2,Ar(Ar(Base(v8),Prop),Base(v7))),DB(1,Ar(Base(v6),Prop)))),Ap(DB(2,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v3),Prop))))),Ap(Ap(Eq(Ar(Base(v2),Prop)),DB(1,Ar(Base(v1),Prop))),DB(0,Ar(Base(v0),Prop))))))))),False)))),False) when basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Forall(Ar(Base(v15),Prop)),Lam(Ar(Base(v14),Prop),Ap(Forall(Ar(Base(v13),Prop)),Lam(Ar(Base(v12),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v11)),Ap(Name(v8,Ar(Ar(Base(v10),Prop),Base(v9))),DB(1,Ar(Base(v7),Prop)))),Ap(Name(v4,Ar(Ar(Base(v6),Prop),Base(v5))),DB(0,Ar(Base(v3),Prop))))),Ap(Ap(Eq(Ar(Base(v2),Prop)),DB(1,Ar(Base(v1),Prop))),DB(0,Ar(Base(v0),Prop)))))))) when basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && eigen__0_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && eigen__0_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 ->
      let f = Name(v4,Ar(Ar(Base(a),Prop),Base(a))) in
      let ao = Ar(Base(a),Prop) in
      let x = DB(1,Base(a)) in
      let y = DB(0,ao) in
      let d = Lam(Base(a),ex ao (conj (eq (Base(a)) x (Ap(f,y))) (neg (Ap(y,x))))) in
      suggstack := SuggCut(Ap(d,Ap(f,d)))::SuggInst(d)::SuggInst(Ap(f,d))::!suggstack;
      -1
  | Ap(Ap(Eq(Base(v19)),Ap(Name(v16,Ar(Ar(Base(v18),Prop),Base(v17))),Lam(Base(v15),Ap(Ap(Imp,Ap(Forall(Ar(Base(v14),Prop)),Lam(Ar(Base(v13),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v12)),DB(1,Base(v11))),Ap(Name(v8,Ar(Ar(Base(v10),Prop),Base(v9))),DB(0,Ar(Base(v7),Prop))))),Ap(DB(0,Ar(Base(v6),Prop)),DB(1,Base(v5))))))),False)))),Ap(Name(v2,Ar(Ar(Base(v4),Prop),Base(v3))),Name(v0,Ar(Base(v1),Prop)))) when basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && eigen__0_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && eigen__0_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && basetp_i_p v3 && eigen__0_p v2 && basetp_i_p v1 && eigen__1_p v0 ->
      suggstack := SuggInst(Name(v0,Ar(Base(a),Prop)))::!suggstack;
      -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Eq(Base(v35)),Ap(Name(v32,Ar(Ar(Base(v34),Prop),Base(v33))),Lam(Base(v31),Ap(Ap(Imp,Ap(Forall(Ar(Base(v30),Prop)),Lam(Ar(Base(v29),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v28)),DB(1,Base(v27))),Ap(Name(v24,Ar(Ar(Base(v26),Prop),Base(v25))),DB(0,Ar(Base(v23),Prop))))),Ap(DB(0,Ar(Base(v22),Prop)),DB(1,Base(v21))))))),False)))),Ap(Name(v18,Ar(Ar(Base(v20),Prop),Base(v19))),Name(v16,Ar(Base(v17),Prop))))),Ap(Name(v14,Ar(Base(v15),Prop)),Ap(Name(v11,Ar(Ar(Base(v13),Prop),Base(v12))),Lam(Base(v10),Ap(Ap(Imp,Ap(Forall(Ar(Base(v9),Prop)),Lam(Ar(Base(v8),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),DB(1,Base(v6))),Ap(Name(v3,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v2),Prop))))),Ap(DB(0,Ar(Base(v1),Prop)),DB(1,Base(v0))))))),False)))))),False) when basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && eigen__0_p v32 && basetp_i_p v31 && basetp_i_p v30 && basetp_i_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && eigen__0_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && eigen__0_p v18 && basetp_i_p v17 && eigen__1_p v16 && basetp_i_p v15 && eigen__1_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__0_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Name(v14,Ar(Base(v15),Prop)),Ap(Name(v11,Ar(Ar(Base(v13),Prop),Base(v12))),Lam(Base(v10),Ap(Ap(Imp,Ap(Forall(Ar(Base(v9),Prop)),Lam(Ar(Base(v8),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),DB(1,Base(v6))),Ap(Name(v3,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v2),Prop))))),Ap(DB(0,Ar(Base(v1),Prop)),DB(1,Base(v0))))))),False))))),False) when basetp_i_p v15 && eigen__1_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__0_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Ar(Base(v22),Prop)),Lam(Base(v21),Ap(Ap(Imp,Ap(Forall(Ar(Base(v20),Prop)),Lam(Ar(Base(v19),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v18)),DB(1,Base(v17))),Ap(Name(v14,Ar(Ar(Base(v16),Prop),Base(v15))),DB(0,Ar(Base(v13),Prop))))),Ap(DB(0,Ar(Base(v12),Prop)),DB(1,Base(v11))))))),False))),Lam(Base(v10),Ap(Ap(Imp,Ap(Forall(Ar(Base(v9),Prop)),Lam(Ar(Base(v8),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),DB(1,Base(v6))),Ap(Name(v3,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v2),Prop))))),Ap(DB(0,Ar(Base(v1),Prop)),DB(1,Base(v0))))))),False)))),False) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && eigen__0_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Eq(Ar(Base(v13),Prop)),Lam(Base(v12),Ap(Ap(Imp,Ap(Forall(Ar(Base(v11),Prop)),Lam(Ar(Base(v10),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v9)),DB(1,Base(v8))),Ap(Name(v5,Ar(Ar(Base(v7),Prop),Base(v6))),DB(0,Ar(Base(v4),Prop))))),Ap(DB(0,Ar(Base(v3),Prop)),DB(1,Base(v2))))))),False))),Name(v0,Ar(Base(v1),Prop))) when basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && eigen__0_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__1_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v33)),Ap(Name(v30,Ar(Ar(Base(v32),Prop),Base(v31))),Lam(Base(v29),Ap(Ap(Imp,Ap(Forall(Ar(Base(v28),Prop)),Lam(Ar(Base(v27),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v26)),DB(1,Base(v25))),Ap(Name(v22,Ar(Ar(Base(v24),Prop),Base(v23))),DB(0,Ar(Base(v21),Prop))))),Ap(DB(0,Ar(Base(v20),Prop)),DB(1,Base(v19))))))),False)))),Ap(Name(v16,Ar(Ar(Base(v18),Prop),Base(v17))),Name(v14,Ar(Base(v15),Prop))))),Ap(Ap(Eq(Ar(Base(v13),Prop)),Lam(Base(v12),Ap(Ap(Imp,Ap(Forall(Ar(Base(v11),Prop)),Lam(Ar(Base(v10),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v9)),DB(1,Base(v8))),Ap(Name(v5,Ar(Ar(Base(v7),Prop),Base(v6))),DB(0,Ar(Base(v4),Prop))))),Ap(DB(0,Ar(Base(v3),Prop)),DB(1,Base(v2))))))),False))),Name(v0,Ar(Base(v1),Prop)))) when basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && eigen__0_p v30 && basetp_i_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && eigen__0_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && eigen__0_p v16 && basetp_i_p v15 && eigen__1_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && eigen__0_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__1_p v0 -> -1
  | Ap(Forall(Base(v14)),Lam(Base(v13),Ap(Ap(Eq(Prop),Ap(Ap(Imp,Ap(Forall(Ar(Base(v12),Prop)),Lam(Ar(Base(v11),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v10)),DB(1,Base(v9))),Ap(Name(v6,Ar(Ar(Base(v8),Prop),Base(v7))),DB(0,Ar(Base(v5),Prop))))),Ap(DB(0,Ar(Base(v4),Prop)),DB(1,Base(v3))))))),False)),Ap(Name(v1,Ar(Base(v2),Prop)),DB(0,Base(v0)))))) when basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && eigen__0_p v6 && basetp_i_p v5 && basetp_i_p v4 && basetp_i_p v3 && basetp_i_p v2 && eigen__1_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Eq(Prop),Ap(Ap(Imp,Ap(Forall(Ar(Base(v51),Prop)),Lam(Ar(Base(v50),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v49)),Ap(Name(v46,Ar(Ar(Base(v48),Prop),Base(v47))),Lam(Base(v45),Ap(Ap(Imp,Ap(Forall(Ar(Base(v44),Prop)),Lam(Ar(Base(v43),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v42)),DB(1,Base(v41))),Ap(Name(v38,Ar(Ar(Base(v40),Prop),Base(v39))),DB(0,Ar(Base(v37),Prop))))),Ap(DB(0,Ar(Base(v36),Prop)),DB(1,Base(v35))))))),False)))),Ap(Name(v32,Ar(Ar(Base(v34),Prop),Base(v33))),DB(0,Ar(Base(v31),Prop))))),Ap(DB(0,Ar(Base(v30),Prop)),Ap(Name(v27,Ar(Ar(Base(v29),Prop),Base(v28))),Lam(Base(v26),Ap(Ap(Imp,Ap(Forall(Ar(Base(v25),Prop)),Lam(Ar(Base(v24),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v23)),DB(1,Base(v22))),Ap(Name(v19,Ar(Ar(Base(v21),Prop),Base(v20))),DB(0,Ar(Base(v18),Prop))))),Ap(DB(0,Ar(Base(v17),Prop)),DB(1,Base(v16))))))),False)))))))),False)),Ap(Name(v14,Ar(Base(v15),Prop)),Ap(Name(v11,Ar(Ar(Base(v13),Prop),Base(v12))),Lam(Base(v10),Ap(Ap(Imp,Ap(Forall(Ar(Base(v9),Prop)),Lam(Ar(Base(v8),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),DB(1,Base(v6))),Ap(Name(v3,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v2),Prop))))),Ap(DB(0,Ar(Base(v1),Prop)),DB(1,Base(v0))))))),False))))) when basetp_i_p v51 && basetp_i_p v50 && basetp_i_p v49 && basetp_i_p v48 && basetp_i_p v47 && eigen__0_p v46 && basetp_i_p v45 && basetp_i_p v44 && basetp_i_p v43 && basetp_i_p v42 && basetp_i_p v41 && basetp_i_p v40 && basetp_i_p v39 && eigen__0_p v38 && basetp_i_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && eigen__0_p v32 && basetp_i_p v31 && basetp_i_p v30 && basetp_i_p v29 && basetp_i_p v28 && eigen__0_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && eigen__0_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && eigen__1_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__0_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Forall(Ar(Base(v33),Prop)),Lam(Ar(Base(v32),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v31)),Ap(Name(v28,Ar(Ar(Base(v30),Prop),Base(v29))),Lam(Base(v27),Ap(Ap(Imp,Ap(Forall(Ar(Base(v26),Prop)),Lam(Ar(Base(v25),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v24)),DB(1,Base(v23))),Ap(Name(v20,Ar(Ar(Base(v22),Prop),Base(v21))),DB(0,Ar(Base(v19),Prop))))),Ap(DB(0,Ar(Base(v18),Prop)),DB(1,Base(v17))))))),False)))),Ap(Name(v14,Ar(Ar(Base(v16),Prop),Base(v15))),DB(0,Ar(Base(v13),Prop))))),Ap(Ap(Eq(Ar(Base(v12),Prop)),Lam(Base(v11),Ap(Ap(Imp,Ap(Forall(Ar(Base(v10),Prop)),Lam(Ar(Base(v9),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v8)),DB(1,Base(v7))),Ap(Name(v4,Ar(Ar(Base(v6),Prop),Base(v5))),DB(0,Ar(Base(v3),Prop))))),Ap(DB(0,Ar(Base(v2),Prop)),DB(1,Base(v1))))))),False))),DB(0,Ar(Base(v0),Prop)))))) when basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && basetp_i_p v30 && basetp_i_p v29 && eigen__0_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && eigen__0_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && basetp_i_p v15 && eigen__0_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && eigen__0_p v4 && basetp_i_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v64)),Ap(Name(v61,Ar(Ar(Base(v63),Prop),Base(v62))),Lam(Base(v60),Ap(Ap(Imp,Ap(Forall(Ar(Base(v59),Prop)),Lam(Ar(Base(v58),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v57)),DB(1,Base(v56))),Ap(Name(v53,Ar(Ar(Base(v55),Prop),Base(v54))),DB(0,Ar(Base(v52),Prop))))),Ap(DB(0,Ar(Base(v51),Prop)),DB(1,Base(v50))))))),False)))),Ap(Name(v47,Ar(Ar(Base(v49),Prop),Base(v48))),Lam(Base(v46),Ap(Ap(Imp,Ap(Forall(Ar(Base(v45),Prop)),Lam(Ar(Base(v44),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v43)),DB(1,Base(v42))),Ap(Name(v39,Ar(Ar(Base(v41),Prop),Base(v40))),DB(0,Ar(Base(v38),Prop))))),Ap(DB(0,Ar(Base(v37),Prop)),DB(1,Base(v36))))))),False))))),Ap(Ap(Imp,Ap(Forall(Ar(Base(v35),Prop)),Lam(Ar(Base(v34),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v33)),Ap(Name(v30,Ar(Ar(Base(v32),Prop),Base(v31))),Lam(Base(v29),Ap(Ap(Imp,Ap(Forall(Ar(Base(v28),Prop)),Lam(Ar(Base(v27),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v26)),DB(1,Base(v25))),Ap(Name(v22,Ar(Ar(Base(v24),Prop),Base(v23))),DB(0,Ar(Base(v21),Prop))))),Ap(DB(0,Ar(Base(v20),Prop)),DB(1,Base(v19))))))),False)))),Ap(Name(v16,Ar(Ar(Base(v18),Prop),Base(v17))),DB(0,Ar(Base(v15),Prop))))),Ap(DB(0,Ar(Base(v14),Prop)),Ap(Name(v11,Ar(Ar(Base(v13),Prop),Base(v12))),Lam(Base(v10),Ap(Ap(Imp,Ap(Forall(Ar(Base(v9),Prop)),Lam(Ar(Base(v8),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),DB(1,Base(v6))),Ap(Name(v3,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v2),Prop))))),Ap(DB(0,Ar(Base(v1),Prop)),DB(1,Base(v0))))))),False)))))))),False)) when basetp_i_p v64 && basetp_i_p v63 && basetp_i_p v62 && eigen__0_p v61 && basetp_i_p v60 && basetp_i_p v59 && basetp_i_p v58 && basetp_i_p v57 && basetp_i_p v56 && basetp_i_p v55 && basetp_i_p v54 && eigen__0_p v53 && basetp_i_p v52 && basetp_i_p v51 && basetp_i_p v50 && basetp_i_p v49 && basetp_i_p v48 && eigen__0_p v47 && basetp_i_p v46 && basetp_i_p v45 && basetp_i_p v44 && basetp_i_p v43 && basetp_i_p v42 && basetp_i_p v41 && basetp_i_p v40 && eigen__0_p v39 && basetp_i_p v38 && basetp_i_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && eigen__0_p v30 && basetp_i_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && eigen__0_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && eigen__0_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__0_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v28)),Ap(Name(v25,Ar(Ar(Base(v27),Prop),Base(v26))),Lam(Base(v24),Ap(Ap(Imp,Ap(Forall(Ar(Base(v23),Prop)),Lam(Ar(Base(v22),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v21)),DB(1,Base(v20))),Ap(Name(v17,Ar(Ar(Base(v19),Prop),Base(v18))),DB(0,Ar(Base(v16),Prop))))),Ap(DB(0,Ar(Base(v15),Prop)),DB(1,Base(v14))))))),False)))),Ap(Name(v11,Ar(Ar(Base(v13),Prop),Base(v12))),Lam(Base(v10),Ap(Ap(Imp,Ap(Forall(Ar(Base(v9),Prop)),Lam(Ar(Base(v8),Prop),Ap(Ap(Imp,Ap(Ap(Eq(Base(v7)),DB(1,Base(v6))),Ap(Name(v3,Ar(Ar(Base(v5),Prop),Base(v4))),DB(0,Ar(Base(v2),Prop))))),Ap(DB(0,Ar(Base(v1),Prop)),DB(1,Base(v0))))))),False))))),False) when basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && eigen__0_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && eigen__0_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && eigen__0_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && eigen__0_p v3 && basetp_i_p v2 && basetp_i_p v1 && basetp_i_p v0 -> -1
  | _ ->
      if !verbose > 9 then (Printf.printf "Looks bad %s\n%s\n" (trm_str m) (trm_ocaml_match m); flush stdout);
      p + 500;;

  (*** Hard Coded Example 5: commutativity of +, induction with 2 subinductions ***)
let induction : (string * string * string) option ref  = ref None;;
let neg_com : (string * string) option ref = ref None;;

let recog_nat_ind m =
  match !induction with
  | Some(_) -> ()
  | None -> 
      match m with
      | Ap(Forall(Ar(Base(v15),Prop)),Lam(Ar(Base(v14),Prop),Ap(Ap(Imp,Ap(DB(0,Ar(Base(v13),Prop)),Name(v11,Base(v12)))),Ap(Ap(Imp,Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Imp,Ap(DB(1,Ar(Base(v8),Prop)),DB(0,Base(v7)))),Ap(DB(1,Ar(Base(v6),Prop)),Ap(Name(v3,Ar(Base(v5),Base(v4))),DB(0,Base(v2)))))))),Ap(Forall(Base(v1)),DB(0,Ar(Base(v0),Prop))))))) when v15 = v14 && v15 = v13 && v15 = v12 && v15 = v10 && v15 = v9 && v15 = v8 && v15 = v7 && v15 = v6 && v15 = v5 && v15 = v4 && v15 = v2 && v15 = v1 && v15 = v0 ->
	  induction := Some(v15,v11,v3)
      | _ -> ();;
	
let recog_neg_com m =
  match !neg_com with
  | Some(_) -> ()
  | None ->
      match !induction with
      | None -> ()
      | Some(nattp,zer,suc) ->
	  let basetp_i_p v = v = nattp in
	  match m with
	  | Ap(Ap(Imp,Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Eq(Base(v14)),Ap(Ap(Name(v10,Ar(Base(v13),Ar(Base(v12),Base(v11)))),Name(v8,Base(v9))),DB(0,Base(v7)))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),DB(0,Base(v2))),Name(v0,Base(v1))))))),False) when basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && basetp_i_p v9 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && v10 = v3 && basetp_i_p v2 && basetp_i_p v1 && v8 = v0 ->
	      let i = Base(nattp) in
	      let io = Ar(i,Prop) in
	      let add x y = Ap(Ap(Name(v10,Ar(i,Ar(i,i))),x),y) in
	      let c = Name(v8,i) in
	      let indp1 = Lam(i,eq i (add c (DB(0,i))) (add (DB(0,i)) c)) in
	      let indp2 = Lam(i,eq i (add (DB(0,i)) (Name(zer,i))) (DB(0,i))) in
	      suggstack := SuggInst(indp1)::SuggInst(indp2)::SuggInst(Name(zer,i))::SuggInst(c)::!suggstack;
	      neg_com := Some(v10,v8)
	  | _ -> ();;

let adapted_process_prop_1_add_com nattp zer suc add c p m =
  let basetp_i_p v = v = nattp in
  let z_p v = v = zer in
  let s_p v = v = suc in
  let a_p v = v = add in
  let eigen__0_p v = v = c in
  let eigen__1_p v =
    try
      match Hashtbl.find eigenvarforh v with
      | Ap(Forall(Base(v25)),Lam(Base(v24),Ap(Ap(Imp,Ap(Ap(Eq(Base(v23)),Ap(Ap(Name(v19,Ar(Base(v22),Ar(Base(v21),Base(v20)))),DB(0,Base(v18))),Name(v16,Base(v17)))),DB(0,Base(v15)))),Ap(Ap(Eq(Base(v14)),Ap(Ap(Name(v10,Ar(Base(v13),Ar(Base(v12),Base(v11)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),DB(0,Base(v6)))),Name(v4,Base(v5)))),Ap(Name(v1,Ar(Base(v3),Base(v2))),DB(0,Base(v0))))))) when basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && a_p v19 && basetp_i_p v18 && basetp_i_p v17 && z_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && a_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && z_p v4 && basetp_i_p v3 && basetp_i_p v2 && s_p v1 && basetp_i_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  let eigen__2_p v =
    try
      match Hashtbl.find eigenvarforh v with
      | Ap(Forall(Base(v37)),Lam(Base(v36),Ap(Ap(Imp,Ap(Ap(Eq(Base(v35)),Ap(Ap(Name(v31,Ar(Base(v34),Ar(Base(v33),Base(v32)))),Name(v29,Base(v30))),DB(0,Base(v28)))),Ap(Ap(Name(v24,Ar(Base(v27),Ar(Base(v26),Base(v25)))),DB(0,Base(v23))),Name(v21,Base(v22))))),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Name(v14,Base(v15))),Ap(Name(v11,Ar(Base(v13),Base(v12))),DB(0,Base(v10))))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Ap(Name(v3,Ar(Base(v5),Base(v4))),DB(0,Base(v2)))),Name(v0,Base(v1))))))) when basetp_i_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && a_p v31 && basetp_i_p v30 && eigen__0_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && a_p v24 && basetp_i_p v23 && basetp_i_p v22 && eigen__0_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && eigen__0_p v14 && basetp_i_p v13 && basetp_i_p v12 && s_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && a_p v6 && basetp_i_p v5 && basetp_i_p v4 && s_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__0_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  let eigen__3_p v =
    try
      match Hashtbl.find eigenvarforh v with
      | Ap(Forall(Base(v49)),Lam(Base(v48),Ap(Ap(Imp,Ap(Ap(Eq(Base(v47)),Ap(Ap(Name(v43,Ar(Base(v46),Ar(Base(v45),Base(v44)))),DB(0,Base(v42))),Ap(Name(v39,Ar(Base(v41),Base(v40))),Name(v37,Base(v38))))),Ap(Name(v34,Ar(Base(v36),Base(v35))),Ap(Ap(Name(v30,Ar(Base(v33),Ar(Base(v32),Base(v31)))),DB(0,Base(v29))),Name(v27,Base(v28)))))),Ap(Ap(Eq(Base(v26)),Ap(Ap(Name(v22,Ar(Base(v25),Ar(Base(v24),Base(v23)))),Ap(Name(v19,Ar(Base(v21),Base(v20))),DB(0,Base(v18)))),Ap(Name(v15,Ar(Base(v17),Base(v16))),Name(v13,Base(v14))))),Ap(Name(v10,Ar(Base(v12),Base(v11))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Ap(Name(v3,Ar(Base(v5),Base(v4))),DB(0,Base(v2)))),Name(v0,Base(v1)))))))) when basetp_i_p v49 && basetp_i_p v48 && basetp_i_p v47 && basetp_i_p v46 && basetp_i_p v45 && basetp_i_p v44 && a_p v43 && basetp_i_p v42 && basetp_i_p v41 && basetp_i_p v40 && s_p v39 && basetp_i_p v38 && eigen__2_p v37 && basetp_i_p v36 && basetp_i_p v35 && s_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && a_p v30 && basetp_i_p v29 && basetp_i_p v28 && eigen__2_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && a_p v22 && basetp_i_p v21 && basetp_i_p v20 && s_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && s_p v15 && basetp_i_p v14 && eigen__2_p v13 && basetp_i_p v12 && basetp_i_p v11 && s_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && a_p v6 && basetp_i_p v5 && basetp_i_p v4 && s_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__2_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  let ok_com_trm_p m =
    let sawa = ref false in
    let scount = ref 0 in
    let zcount = ref 0 in
    let rec ok_com_trm_p_r m =
      match m with
      | Ap(Ap(Name(v,_),m1),m2) when a_p v && not !sawa ->
	  sawa := true;
	  ok_com_trm_p_r m1 && ok_com_trm_p_r m2
      | Ap(Name(v,_),m1) when s_p v && !scount < 3 ->
	  incr scount;
	  ok_com_trm_p_r m1
      | Name(v,_) when z_p v && !zcount < 3 ->
	  incr zcount;
	  true
      | Name(v,_) -> true
      | _ ->
	  false
    in
    ok_com_trm_p_r m
  in
  match m with
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v47)),Ap(Ap(Name(v43,Ar(Base(v46),Ar(Base(v45),Base(v44)))),Name(v41,Base(v42))),Name(v39,Base(v40)))),Name(v37,Base(v38)))),Ap(Ap(Imp,Ap(Forall(Base(v36)),Lam(Base(v35),Ap(Ap(Imp,Ap(Ap(Eq(Base(v34)),Ap(Ap(Name(v30,Ar(Base(v33),Ar(Base(v32),Base(v31)))),DB(0,Base(v29))),Name(v27,Base(v28)))),DB(0,Base(v26)))),Ap(Ap(Eq(Base(v25)),Ap(Ap(Name(v21,Ar(Base(v24),Ar(Base(v23),Base(v22)))),Ap(Name(v18,Ar(Base(v20),Base(v19))),DB(0,Base(v17)))),Name(v15,Base(v16)))),Ap(Name(v12,Ar(Base(v14),Base(v13))),DB(0,Base(v11)))))))),Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Eq(Base(v8)),Ap(Ap(Name(v4,Ar(Base(v7),Ar(Base(v6),Base(v5)))),DB(0,Base(v3))),Name(v1,Base(v2)))),DB(0,Base(v0))))))) when basetp_i_p v47 && basetp_i_p v46 && basetp_i_p v45 && basetp_i_p v44 && a_p v43 && basetp_i_p v42 && z_p v41 && basetp_i_p v40 && z_p v39 && basetp_i_p v38 && z_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && a_p v30 && basetp_i_p v29 && basetp_i_p v28 && z_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && a_p v21 && basetp_i_p v20 && basetp_i_p v19 && s_p v18 && basetp_i_p v17 && basetp_i_p v16 && z_p v15 && basetp_i_p v14 && basetp_i_p v13 && s_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && a_p v4 && basetp_i_p v3 && basetp_i_p v2 && z_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v71)),Ap(Ap(Name(v67,Ar(Base(v70),Ar(Base(v69),Base(v68)))),Name(v65,Base(v66))),Name(v63,Base(v64)))),Ap(Ap(Name(v59,Ar(Base(v62),Ar(Base(v61),Base(v60)))),Name(v57,Base(v58))),Name(v55,Base(v56))))),Ap(Ap(Imp,Ap(Forall(Base(v54)),Lam(Base(v53),Ap(Ap(Imp,Ap(Ap(Eq(Base(v52)),Ap(Ap(Name(v48,Ar(Base(v51),Ar(Base(v50),Base(v49)))),Name(v46,Base(v47))),DB(0,Base(v45)))),Ap(Ap(Name(v41,Ar(Base(v44),Ar(Base(v43),Base(v42)))),DB(0,Base(v40))),Name(v38,Base(v39))))),Ap(Ap(Eq(Base(v37)),Ap(Ap(Name(v33,Ar(Base(v36),Ar(Base(v35),Base(v34)))),Name(v31,Base(v32))),Ap(Name(v28,Ar(Base(v30),Base(v29))),DB(0,Base(v27))))),Ap(Ap(Name(v23,Ar(Base(v26),Ar(Base(v25),Base(v24)))),Ap(Name(v20,Ar(Base(v22),Base(v21))),DB(0,Base(v19)))),Name(v17,Base(v18)))))))),Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Eq(Base(v14)),Ap(Ap(Name(v10,Ar(Base(v13),Ar(Base(v12),Base(v11)))),Name(v8,Base(v9))),DB(0,Base(v7)))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),DB(0,Base(v2))),Name(v0,Base(v1)))))))) when basetp_i_p v71 && basetp_i_p v70 && basetp_i_p v69 && basetp_i_p v68 && a_p v67 && basetp_i_p v66 && eigen__0_p v65 && basetp_i_p v64 && z_p v63 && basetp_i_p v62 && basetp_i_p v61 && basetp_i_p v60 && a_p v59 && basetp_i_p v58 && z_p v57 && basetp_i_p v56 && eigen__0_p v55 && basetp_i_p v54 && basetp_i_p v53 && basetp_i_p v52 && basetp_i_p v51 && basetp_i_p v50 && basetp_i_p v49 && a_p v48 && basetp_i_p v47 && eigen__0_p v46 && basetp_i_p v45 && basetp_i_p v44 && basetp_i_p v43 && basetp_i_p v42 && a_p v41 && basetp_i_p v40 && basetp_i_p v39 && eigen__0_p v38 && basetp_i_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && a_p v33 && basetp_i_p v32 && eigen__0_p v31 && basetp_i_p v30 && basetp_i_p v29 && s_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && a_p v23 && basetp_i_p v22 && basetp_i_p v21 && s_p v20 && basetp_i_p v19 && basetp_i_p v18 && eigen__0_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && a_p v10 && basetp_i_p v9 && eigen__0_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__0_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v54)),Lam(Base(v53),Ap(Ap(Imp,Ap(Ap(Eq(Base(v52)),Ap(Ap(Name(v48,Ar(Base(v51),Ar(Base(v50),Base(v49)))),Name(v46,Base(v47))),DB(0,Base(v45)))),Ap(Ap(Name(v41,Ar(Base(v44),Ar(Base(v43),Base(v42)))),DB(0,Base(v40))),Name(v38,Base(v39))))),Ap(Ap(Eq(Base(v37)),Ap(Ap(Name(v33,Ar(Base(v36),Ar(Base(v35),Base(v34)))),Name(v31,Base(v32))),Ap(Name(v28,Ar(Base(v30),Base(v29))),DB(0,Base(v27))))),Ap(Ap(Name(v23,Ar(Base(v26),Ar(Base(v25),Base(v24)))),Ap(Name(v20,Ar(Base(v22),Base(v21))),DB(0,Base(v19)))),Name(v17,Base(v18)))))))),Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Eq(Base(v14)),Ap(Ap(Name(v10,Ar(Base(v13),Ar(Base(v12),Base(v11)))),Name(v8,Base(v9))),DB(0,Base(v7)))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),DB(0,Base(v2))),Name(v0,Base(v1))))))) when basetp_i_p v54 && basetp_i_p v53 && basetp_i_p v52 && basetp_i_p v51 && basetp_i_p v50 && basetp_i_p v49 && a_p v48 && basetp_i_p v47 && eigen__0_p v46 && basetp_i_p v45 && basetp_i_p v44 && basetp_i_p v43 && basetp_i_p v42 && a_p v41 && basetp_i_p v40 && basetp_i_p v39 && eigen__0_p v38 && basetp_i_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && a_p v33 && basetp_i_p v32 && eigen__0_p v31 && basetp_i_p v30 && basetp_i_p v29 && s_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && a_p v23 && basetp_i_p v22 && basetp_i_p v21 && s_p v20 && basetp_i_p v19 && basetp_i_p v18 && eigen__0_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && a_p v10 && basetp_i_p v9 && eigen__0_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__0_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v36)),Lam(Base(v35),Ap(Ap(Imp,Ap(Ap(Eq(Base(v34)),Ap(Ap(Name(v30,Ar(Base(v33),Ar(Base(v32),Base(v31)))),DB(0,Base(v29))),Name(v27,Base(v28)))),DB(0,Base(v26)))),Ap(Ap(Eq(Base(v25)),Ap(Ap(Name(v21,Ar(Base(v24),Ar(Base(v23),Base(v22)))),Ap(Name(v18,Ar(Base(v20),Base(v19))),DB(0,Base(v17)))),Name(v15,Base(v16)))),Ap(Name(v12,Ar(Base(v14),Base(v13))),DB(0,Base(v11)))))))),Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Eq(Base(v8)),Ap(Ap(Name(v4,Ar(Base(v7),Ar(Base(v6),Base(v5)))),DB(0,Base(v3))),Name(v1,Base(v2)))),DB(0,Base(v0)))))) when basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && a_p v30 && basetp_i_p v29 && basetp_i_p v28 && z_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && a_p v21 && basetp_i_p v20 && basetp_i_p v19 && s_p v18 && basetp_i_p v17 && basetp_i_p v16 && z_p v15 && basetp_i_p v14 && basetp_i_p v13 && s_p v12 && basetp_i_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && a_p v4 && basetp_i_p v3 && basetp_i_p v2 && z_p v1 && basetp_i_p v0 -> -1
  | Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Ap(Name(v13,Ar(Base(v15),Base(v14))),Name(v11,Base(v12)))),DB(0,Base(v10)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),Name(v1,Base(v2))),DB(0,Base(v0))))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && s_p v13 && basetp_i_p v12 && z_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && z_p v1 && basetp_i_p v0 -> -1
  | Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Ap(Name(v13,Ar(Base(v15),Base(v14))),Name(v11,Base(v12)))),DB(0,Base(v10)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),Name(v1,Base(v2))),DB(0,Base(v0))))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && s_p v13 && basetp_i_p v12 && eigen__0_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && eigen__0_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v25)),Lam(Base(v24),Ap(Ap(Imp,Ap(Ap(Eq(Base(v23)),Ap(Ap(Name(v19,Ar(Base(v22),Ar(Base(v21),Base(v20)))),DB(0,Base(v18))),Name(v16,Base(v17)))),DB(0,Base(v15)))),Ap(Ap(Eq(Base(v14)),Ap(Ap(Name(v10,Ar(Base(v13),Ar(Base(v12),Base(v11)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),DB(0,Base(v6)))),Name(v4,Base(v5)))),Ap(Name(v1,Ar(Base(v3),Base(v2))),DB(0,Base(v0)))))))),False) when basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && a_p v19 && basetp_i_p v18 && basetp_i_p v17 && z_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && a_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && z_p v4 && basetp_i_p v3 && basetp_i_p v2 && s_p v1 && basetp_i_p v0 -> -1
  | Ap(Forall(Base(v10)),Lam(Base(v9),Ap(Ap(Eq(Base(v8)),Ap(Ap(Name(v4,Ar(Base(v7),Ar(Base(v6),Base(v5)))),DB(0,Base(v3))),Name(v1,Base(v2)))),DB(0,Base(v0))))) when basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && a_p v4 && basetp_i_p v3 && basetp_i_p v2 && z_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Eq(Base(v27)),Ap(Ap(Name(v23,Ar(Base(v26),Ar(Base(v25),Base(v24)))),Name(v21,Base(v22))),Name(v19,Base(v20)))),Name(v17,Base(v18)))),Ap(Ap(Eq(Base(v16)),Ap(Ap(Name(v12,Ar(Base(v15),Ar(Base(v14),Base(v13)))),Ap(Name(v9,Ar(Base(v11),Base(v10))),Name(v7,Base(v8)))),Name(v5,Base(v6)))),Ap(Name(v2,Ar(Base(v4),Base(v3))),Name(v0,Base(v1)))))),False) when basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && a_p v23 && basetp_i_p v22 && eigen__1_p v21 && basetp_i_p v20 && z_p v19 && basetp_i_p v18 && eigen__1_p v17 && basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && a_p v12 && basetp_i_p v11 && basetp_i_p v10 && s_p v9 && basetp_i_p v8 && eigen__1_p v7 && basetp_i_p v6 && z_p v5 && basetp_i_p v4 && basetp_i_p v3 && s_p v2 && basetp_i_p v1 && eigen__1_p v0 ->
      suggstack := SuggInst(Name(v0,Base(v1)))::!suggstack;
      -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Eq(Base(v39)),Ap(Ap(Name(v35,Ar(Base(v38),Ar(Base(v37),Base(v36)))),Name(v33,Base(v34))),Name(v31,Base(v32)))),Ap(Ap(Name(v27,Ar(Base(v30),Ar(Base(v29),Base(v28)))),Name(v25,Base(v26))),Name(v23,Base(v24))))),Ap(Ap(Eq(Base(v22)),Ap(Ap(Name(v18,Ar(Base(v21),Ar(Base(v20),Base(v19)))),Name(v16,Base(v17))),Ap(Name(v13,Ar(Base(v15),Base(v14))),Name(v11,Base(v12))))),Ap(Ap(Name(v7,Ar(Base(v10),Ar(Base(v9),Base(v8)))),Ap(Name(v4,Ar(Base(v6),Base(v5))),Name(v2,Base(v3)))),Name(v0,Base(v1)))))),False) when basetp_i_p v39 && basetp_i_p v38 && basetp_i_p v37 && basetp_i_p v36 && a_p v35 && basetp_i_p v34 && eigen__0_p v33 && basetp_i_p v32 && eigen__2_p v31 && basetp_i_p v30 && basetp_i_p v29 && basetp_i_p v28 && a_p v27 && basetp_i_p v26 && eigen__2_p v25 && basetp_i_p v24 && eigen__0_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && a_p v18 && basetp_i_p v17 && eigen__0_p v16 && basetp_i_p v15 && basetp_i_p v14 && s_p v13 && basetp_i_p v12 && eigen__2_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && a_p v7 && basetp_i_p v6 && basetp_i_p v5 && s_p v4 && basetp_i_p v3 && eigen__2_p v2 && basetp_i_p v1 && eigen__0_p v0 ->
      let i = Base(nattp) in
      let d = Name(v11,i) in
      let addop x y = Ap(Ap(Name(add,Ar(i,Ar(i,i))),x),y) in
      let sd = Ap(Name(suc,Ar(i,i)),d) in
      let indp3 = Lam(i,eq i (addop (DB(0,i)) sd) (Ap(Name(suc,Ar(i,i)),addop (DB(0,i)) d))) in
      suggstack := SuggInst(indp3)::SuggInst(d)::SuggInst(sd)::!suggstack;
      -1
  | Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Ap(Name(v13,Ar(Base(v15),Base(v14))),Name(v11,Base(v12)))),DB(0,Base(v10)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),Name(v1,Base(v2))),DB(0,Base(v0))))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && s_p v13 && basetp_i_p v12 && eigen__1_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && eigen__1_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v37)),Lam(Base(v36),Ap(Ap(Imp,Ap(Ap(Eq(Base(v35)),Ap(Ap(Name(v31,Ar(Base(v34),Ar(Base(v33),Base(v32)))),Name(v29,Base(v30))),DB(0,Base(v28)))),Ap(Ap(Name(v24,Ar(Base(v27),Ar(Base(v26),Base(v25)))),DB(0,Base(v23))),Name(v21,Base(v22))))),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Name(v14,Base(v15))),Ap(Name(v11,Ar(Base(v13),Base(v12))),DB(0,Base(v10))))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Ap(Name(v3,Ar(Base(v5),Base(v4))),DB(0,Base(v2)))),Name(v0,Base(v1)))))))),False) when basetp_i_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && a_p v31 && basetp_i_p v30 && eigen__0_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && a_p v24 && basetp_i_p v23 && basetp_i_p v22 && eigen__0_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && eigen__0_p v14 && basetp_i_p v13 && basetp_i_p v12 && s_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && a_p v6 && basetp_i_p v5 && basetp_i_p v4 && s_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__0_p v0 -> -1
  | Ap(Forall(Base(v16)),Lam(Base(v15),Ap(Ap(Eq(Base(v14)),Ap(Ap(Name(v10,Ar(Base(v13),Ar(Base(v12),Base(v11)))),Name(v8,Base(v9))),DB(0,Base(v7)))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),DB(0,Base(v2))),Name(v0,Base(v1)))))) when basetp_i_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && a_p v10 && basetp_i_p v9 && eigen__0_p v8 && basetp_i_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__0_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v95)),Ap(Ap(Name(v91,Ar(Base(v94),Ar(Base(v93),Base(v92)))),Name(v89,Base(v90))),Ap(Name(v86,Ar(Base(v88),Base(v87))),Name(v84,Base(v85))))),Ap(Name(v81,Ar(Base(v83),Base(v82))),Ap(Ap(Name(v77,Ar(Base(v80),Ar(Base(v79),Base(v78)))),Name(v75,Base(v76))),Name(v73,Base(v74)))))),Ap(Ap(Imp,Ap(Forall(Base(v72)),Lam(Base(v71),Ap(Ap(Imp,Ap(Ap(Eq(Base(v70)),Ap(Ap(Name(v66,Ar(Base(v69),Ar(Base(v68),Base(v67)))),DB(0,Base(v65))),Ap(Name(v62,Ar(Base(v64),Base(v63))),Name(v60,Base(v61))))),Ap(Name(v57,Ar(Base(v59),Base(v58))),Ap(Ap(Name(v53,Ar(Base(v56),Ar(Base(v55),Base(v54)))),DB(0,Base(v52))),Name(v50,Base(v51)))))),Ap(Ap(Eq(Base(v49)),Ap(Ap(Name(v45,Ar(Base(v48),Ar(Base(v47),Base(v46)))),Ap(Name(v42,Ar(Base(v44),Base(v43))),DB(0,Base(v41)))),Ap(Name(v38,Ar(Base(v40),Base(v39))),Name(v36,Base(v37))))),Ap(Name(v33,Ar(Base(v35),Base(v34))),Ap(Ap(Name(v29,Ar(Base(v32),Ar(Base(v31),Base(v30)))),Ap(Name(v26,Ar(Base(v28),Base(v27))),DB(0,Base(v25)))),Name(v23,Base(v24))))))))),Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),DB(0,Base(v15))),Ap(Name(v12,Ar(Base(v14),Base(v13))),Name(v10,Base(v11))))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),DB(0,Base(v2))),Name(v0,Base(v1))))))))) when basetp_i_p v95 && basetp_i_p v94 && basetp_i_p v93 && basetp_i_p v92 && a_p v91 && basetp_i_p v90 && z_p v89 && basetp_i_p v88 && basetp_i_p v87 && s_p v86 && basetp_i_p v85 && eigen__2_p v84 && basetp_i_p v83 && basetp_i_p v82 && s_p v81 && basetp_i_p v80 && basetp_i_p v79 && basetp_i_p v78 && a_p v77 && basetp_i_p v76 && z_p v75 && basetp_i_p v74 && eigen__2_p v73 && basetp_i_p v72 && basetp_i_p v71 && basetp_i_p v70 && basetp_i_p v69 && basetp_i_p v68 && basetp_i_p v67 && a_p v66 && basetp_i_p v65 && basetp_i_p v64 && basetp_i_p v63 && s_p v62 && basetp_i_p v61 && eigen__2_p v60 && basetp_i_p v59 && basetp_i_p v58 && s_p v57 && basetp_i_p v56 && basetp_i_p v55 && basetp_i_p v54 && a_p v53 && basetp_i_p v52 && basetp_i_p v51 && eigen__2_p v50 && basetp_i_p v49 && basetp_i_p v48 && basetp_i_p v47 && basetp_i_p v46 && a_p v45 && basetp_i_p v44 && basetp_i_p v43 && s_p v42 && basetp_i_p v41 && basetp_i_p v40 && basetp_i_p v39 && s_p v38 && basetp_i_p v37 && eigen__2_p v36 && basetp_i_p v35 && basetp_i_p v34 && s_p v33 && basetp_i_p v32 && basetp_i_p v31 && basetp_i_p v30 && a_p v29 && basetp_i_p v28 && basetp_i_p v27 && s_p v26 && basetp_i_p v25 && basetp_i_p v24 && eigen__2_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && s_p v12 && basetp_i_p v11 && eigen__2_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__2_p v0 -> -1
  | Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Ap(Name(v13,Ar(Base(v15),Base(v14))),Name(v11,Base(v12)))),DB(0,Base(v10)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),Name(v1,Base(v2))),DB(0,Base(v0))))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && s_p v13 && basetp_i_p v12 && eigen__2_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && eigen__2_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v72)),Lam(Base(v71),Ap(Ap(Imp,Ap(Ap(Eq(Base(v70)),Ap(Ap(Name(v66,Ar(Base(v69),Ar(Base(v68),Base(v67)))),DB(0,Base(v65))),Ap(Name(v62,Ar(Base(v64),Base(v63))),Name(v60,Base(v61))))),Ap(Name(v57,Ar(Base(v59),Base(v58))),Ap(Ap(Name(v53,Ar(Base(v56),Ar(Base(v55),Base(v54)))),DB(0,Base(v52))),Name(v50,Base(v51)))))),Ap(Ap(Eq(Base(v49)),Ap(Ap(Name(v45,Ar(Base(v48),Ar(Base(v47),Base(v46)))),Ap(Name(v42,Ar(Base(v44),Base(v43))),DB(0,Base(v41)))),Ap(Name(v38,Ar(Base(v40),Base(v39))),Name(v36,Base(v37))))),Ap(Name(v33,Ar(Base(v35),Base(v34))),Ap(Ap(Name(v29,Ar(Base(v32),Ar(Base(v31),Base(v30)))),Ap(Name(v26,Ar(Base(v28),Base(v27))),DB(0,Base(v25)))),Name(v23,Base(v24))))))))),Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),DB(0,Base(v15))),Ap(Name(v12,Ar(Base(v14),Base(v13))),Name(v10,Base(v11))))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),DB(0,Base(v2))),Name(v0,Base(v1)))))))) when basetp_i_p v72 && basetp_i_p v71 && basetp_i_p v70 && basetp_i_p v69 && basetp_i_p v68 && basetp_i_p v67 && a_p v66 && basetp_i_p v65 && basetp_i_p v64 && basetp_i_p v63 && s_p v62 && basetp_i_p v61 && eigen__2_p v60 && basetp_i_p v59 && basetp_i_p v58 && s_p v57 && basetp_i_p v56 && basetp_i_p v55 && basetp_i_p v54 && a_p v53 && basetp_i_p v52 && basetp_i_p v51 && eigen__2_p v50 && basetp_i_p v49 && basetp_i_p v48 && basetp_i_p v47 && basetp_i_p v46 && a_p v45 && basetp_i_p v44 && basetp_i_p v43 && s_p v42 && basetp_i_p v41 && basetp_i_p v40 && basetp_i_p v39 && s_p v38 && basetp_i_p v37 && eigen__2_p v36 && basetp_i_p v35 && basetp_i_p v34 && s_p v33 && basetp_i_p v32 && basetp_i_p v31 && basetp_i_p v30 && a_p v29 && basetp_i_p v28 && basetp_i_p v27 && s_p v26 && basetp_i_p v25 && basetp_i_p v24 && eigen__2_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && s_p v12 && basetp_i_p v11 && eigen__2_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__2_p v0 -> -1
  | Ap(Ap(Imp,Ap(Forall(Base(v49)),Lam(Base(v48),Ap(Ap(Imp,Ap(Ap(Eq(Base(v47)),Ap(Ap(Name(v43,Ar(Base(v46),Ar(Base(v45),Base(v44)))),DB(0,Base(v42))),Ap(Name(v39,Ar(Base(v41),Base(v40))),Name(v37,Base(v38))))),Ap(Name(v34,Ar(Base(v36),Base(v35))),Ap(Ap(Name(v30,Ar(Base(v33),Ar(Base(v32),Base(v31)))),DB(0,Base(v29))),Name(v27,Base(v28)))))),Ap(Ap(Eq(Base(v26)),Ap(Ap(Name(v22,Ar(Base(v25),Ar(Base(v24),Base(v23)))),Ap(Name(v19,Ar(Base(v21),Base(v20))),DB(0,Base(v18)))),Ap(Name(v15,Ar(Base(v17),Base(v16))),Name(v13,Base(v14))))),Ap(Name(v10,Ar(Base(v12),Base(v11))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Ap(Name(v3,Ar(Base(v5),Base(v4))),DB(0,Base(v2)))),Name(v0,Base(v1))))))))),False) when basetp_i_p v49 && basetp_i_p v48 && basetp_i_p v47 && basetp_i_p v46 && basetp_i_p v45 && basetp_i_p v44 && a_p v43 && basetp_i_p v42 && basetp_i_p v41 && basetp_i_p v40 && s_p v39 && basetp_i_p v38 && eigen__2_p v37 && basetp_i_p v36 && basetp_i_p v35 && s_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && a_p v30 && basetp_i_p v29 && basetp_i_p v28 && eigen__2_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && a_p v22 && basetp_i_p v21 && basetp_i_p v20 && s_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && s_p v15 && basetp_i_p v14 && eigen__2_p v13 && basetp_i_p v12 && basetp_i_p v11 && s_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && a_p v6 && basetp_i_p v5 && basetp_i_p v4 && s_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__2_p v0 -> -1
  | Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),DB(0,Base(v15))),Ap(Name(v12,Ar(Base(v14),Base(v13))),Name(v10,Base(v11))))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),DB(0,Base(v2))),Name(v0,Base(v1))))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && s_p v12 && basetp_i_p v11 && eigen__2_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__2_p v0 -> -1
  | Ap(Ap(Imp,Ap(Ap(Imp,Ap(Ap(Eq(Base(v51)),Ap(Ap(Name(v47,Ar(Base(v50),Ar(Base(v49),Base(v48)))),Name(v45,Base(v46))),Ap(Name(v42,Ar(Base(v44),Base(v43))),Name(v40,Base(v41))))),Ap(Name(v37,Ar(Base(v39),Base(v38))),Ap(Ap(Name(v33,Ar(Base(v36),Ar(Base(v35),Base(v34)))),Name(v31,Base(v32))),Name(v29,Base(v30)))))),Ap(Ap(Eq(Base(v28)),Ap(Ap(Name(v24,Ar(Base(v27),Ar(Base(v26),Base(v25)))),Ap(Name(v21,Ar(Base(v23),Base(v22))),Name(v19,Base(v20)))),Ap(Name(v16,Ar(Base(v18),Base(v17))),Name(v14,Base(v15))))),Ap(Name(v11,Ar(Base(v13),Base(v12))),Ap(Ap(Name(v7,Ar(Base(v10),Ar(Base(v9),Base(v8)))),Ap(Name(v4,Ar(Base(v6),Base(v5))),Name(v2,Base(v3)))),Name(v0,Base(v1))))))),False) when basetp_i_p v51 && basetp_i_p v50 && basetp_i_p v49 && basetp_i_p v48 && a_p v47 && basetp_i_p v46 && eigen__3_p v45 && basetp_i_p v44 && basetp_i_p v43 && s_p v42 && basetp_i_p v41 && eigen__2_p v40 && basetp_i_p v39 && basetp_i_p v38 && s_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && a_p v33 && basetp_i_p v32 && eigen__3_p v31 && basetp_i_p v30 && eigen__2_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && a_p v24 && basetp_i_p v23 && basetp_i_p v22 && s_p v21 && basetp_i_p v20 && eigen__3_p v19 && basetp_i_p v18 && basetp_i_p v17 && s_p v16 && basetp_i_p v15 && eigen__2_p v14 && basetp_i_p v13 && basetp_i_p v12 && s_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && a_p v7 && basetp_i_p v6 && basetp_i_p v5 && s_p v4 && basetp_i_p v3 && eigen__3_p v2 && basetp_i_p v1 && eigen__2_p v0 ->
      suggstack := SuggInst(Name(v45,Base(nattp)))::!suggstack;
      -1
  | Ap(Forall(Base(v22)),Lam(Base(v21),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Ap(Name(v13,Ar(Base(v15),Base(v14))),Name(v11,Base(v12)))),DB(0,Base(v10)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),Ap(Ap(Name(v3,Ar(Base(v6),Ar(Base(v5),Base(v4)))),Name(v1,Base(v2))),DB(0,Base(v0))))))) when basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && basetp_i_p v14 && s_p v13 && basetp_i_p v12 && eigen__3_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && basetp_i_p v4 && a_p v3 && basetp_i_p v2 && eigen__3_p v1 && basetp_i_p v0 -> -1
  | Ap(Ap(Eq(Base(v)),m1),m2) when basetp_i_p v && ok_com_trm_p m1 && ok_com_trm_p m2 -> -1
  | Ap(Ap(Imp,Ap(Ap(Eq(Base(v)),m1),m2)),False) when basetp_i_p v && ok_com_trm_p m1 && ok_com_trm_p m2 -> -1
  | _ ->
      if !verbose > 9 then (Printf.printf "Looks bad %s\n%s\n" (trm_str m) (trm_ocaml_match m); flush stdout);
      p + 500;;

let adapted_newinst_add_com nattp zer suc add c p mtp m =
  let basetp_i_p v = v = nattp in
  let z_p v = v = zer in
  let s_p v = v = suc in
  let a_p v = v = add in
  let eigen__0_p v = v = c in
  if mtp = Base(nattp) then
    begin
      match m with
      | Name(_,_) -> -1
      | Ap(Name(f,_),Name(v,_)) when s_p f -> -1
      | _ -> 500
    end
  else
    p;;
	
let adapted_confront_add_com nattp zer suc add c p a m1 m2 n1 n2 =
  let basetp_i_p v = v = nattp in
  let z_p v = v = zer in
  let s_p v = v = suc in
  let a_p v = v = add in
  let eigen__0_p v = v = c in
  let eigen__1_p v =
    try
      match Hashtbl.find eigenvarforh v with
      | Ap(Forall(Base(v25)),Lam(Base(v24),Ap(Ap(Imp,Ap(Ap(Eq(Base(v23)),Ap(Ap(Name(v19,Ar(Base(v22),Ar(Base(v21),Base(v20)))),DB(0,Base(v18))),Name(v16,Base(v17)))),DB(0,Base(v15)))),Ap(Ap(Eq(Base(v14)),Ap(Ap(Name(v10,Ar(Base(v13),Ar(Base(v12),Base(v11)))),Ap(Name(v7,Ar(Base(v9),Base(v8))),DB(0,Base(v6)))),Name(v4,Base(v5)))),Ap(Name(v1,Ar(Base(v3),Base(v2))),DB(0,Base(v0))))))) when basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && basetp_i_p v22 && basetp_i_p v21 && basetp_i_p v20 && a_p v19 && basetp_i_p v18 && basetp_i_p v17 && z_p v16 && basetp_i_p v15 && basetp_i_p v14 && basetp_i_p v13 && basetp_i_p v12 && basetp_i_p v11 && a_p v10 && basetp_i_p v9 && basetp_i_p v8 && s_p v7 && basetp_i_p v6 && basetp_i_p v5 && z_p v4 && basetp_i_p v3 && basetp_i_p v2 && s_p v1 && basetp_i_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  let eigen__2_p v =
    try
      match Hashtbl.find eigenvarforh v with
      | Ap(Forall(Base(v37)),Lam(Base(v36),Ap(Ap(Imp,Ap(Ap(Eq(Base(v35)),Ap(Ap(Name(v31,Ar(Base(v34),Ar(Base(v33),Base(v32)))),Name(v29,Base(v30))),DB(0,Base(v28)))),Ap(Ap(Name(v24,Ar(Base(v27),Ar(Base(v26),Base(v25)))),DB(0,Base(v23))),Name(v21,Base(v22))))),Ap(Ap(Eq(Base(v20)),Ap(Ap(Name(v16,Ar(Base(v19),Ar(Base(v18),Base(v17)))),Name(v14,Base(v15))),Ap(Name(v11,Ar(Base(v13),Base(v12))),DB(0,Base(v10))))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Ap(Name(v3,Ar(Base(v5),Base(v4))),DB(0,Base(v2)))),Name(v0,Base(v1))))))) when basetp_i_p v37 && basetp_i_p v36 && basetp_i_p v35 && basetp_i_p v34 && basetp_i_p v33 && basetp_i_p v32 && a_p v31 && basetp_i_p v30 && eigen__0_p v29 && basetp_i_p v28 && basetp_i_p v27 && basetp_i_p v26 && basetp_i_p v25 && a_p v24 && basetp_i_p v23 && basetp_i_p v22 && eigen__0_p v21 && basetp_i_p v20 && basetp_i_p v19 && basetp_i_p v18 && basetp_i_p v17 && a_p v16 && basetp_i_p v15 && eigen__0_p v14 && basetp_i_p v13 && basetp_i_p v12 && s_p v11 && basetp_i_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && a_p v6 && basetp_i_p v5 && basetp_i_p v4 && s_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__0_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  let eigen__3_p v =
    try
      match Hashtbl.find eigenvarforh v with
      | Ap(Forall(Base(v49)),Lam(Base(v48),Ap(Ap(Imp,Ap(Ap(Eq(Base(v47)),Ap(Ap(Name(v43,Ar(Base(v46),Ar(Base(v45),Base(v44)))),DB(0,Base(v42))),Ap(Name(v39,Ar(Base(v41),Base(v40))),Name(v37,Base(v38))))),Ap(Name(v34,Ar(Base(v36),Base(v35))),Ap(Ap(Name(v30,Ar(Base(v33),Ar(Base(v32),Base(v31)))),DB(0,Base(v29))),Name(v27,Base(v28)))))),Ap(Ap(Eq(Base(v26)),Ap(Ap(Name(v22,Ar(Base(v25),Ar(Base(v24),Base(v23)))),Ap(Name(v19,Ar(Base(v21),Base(v20))),DB(0,Base(v18)))),Ap(Name(v15,Ar(Base(v17),Base(v16))),Name(v13,Base(v14))))),Ap(Name(v10,Ar(Base(v12),Base(v11))),Ap(Ap(Name(v6,Ar(Base(v9),Ar(Base(v8),Base(v7)))),Ap(Name(v3,Ar(Base(v5),Base(v4))),DB(0,Base(v2)))),Name(v0,Base(v1)))))))) when basetp_i_p v49 && basetp_i_p v48 && basetp_i_p v47 && basetp_i_p v46 && basetp_i_p v45 && basetp_i_p v44 && a_p v43 && basetp_i_p v42 && basetp_i_p v41 && basetp_i_p v40 && s_p v39 && basetp_i_p v38 && eigen__2_p v37 && basetp_i_p v36 && basetp_i_p v35 && s_p v34 && basetp_i_p v33 && basetp_i_p v32 && basetp_i_p v31 && a_p v30 && basetp_i_p v29 && basetp_i_p v28 && eigen__2_p v27 && basetp_i_p v26 && basetp_i_p v25 && basetp_i_p v24 && basetp_i_p v23 && a_p v22 && basetp_i_p v21 && basetp_i_p v20 && s_p v19 && basetp_i_p v18 && basetp_i_p v17 && basetp_i_p v16 && s_p v15 && basetp_i_p v14 && eigen__2_p v13 && basetp_i_p v12 && basetp_i_p v11 && s_p v10 && basetp_i_p v9 && basetp_i_p v8 && basetp_i_p v7 && a_p v6 && basetp_i_p v5 && basetp_i_p v4 && s_p v3 && basetp_i_p v2 && basetp_i_p v1 && eigen__2_p v0 -> true
      | _ -> false
    with Not_found -> false
  in
  let ok_com_trm_p m =
    let sawa = ref false in
    let scount = ref 0 in
    let zcount = ref 0 in
    let rec ok_com_trm_p_r m =
      match m with
      | Ap(Ap(Name(v,_),m1),m2) when a_p v && not !sawa ->
	  sawa := true;
	  ok_com_trm_p_r m1 && ok_com_trm_p_r m2
      | Ap(Name(v,_),m1) when s_p v && !scount < 3 ->
	  incr scount;
	  ok_com_trm_p_r m1
      | Name(v,_) when z_p v && !zcount < 3 ->
	  incr zcount;
	  true
      | Name(v,_) -> true
      | _ ->
	  false
    in
    ok_com_trm_p_r m
  in
  if ok_com_trm_p m1 && ok_com_trm_p m2 && ok_com_trm_p n1 && ok_com_trm_p n2 then
    if m1 = n1 || m1 = n2 || m2 = n1 || m2 = n2 then
      -1
    else
      20
  else
    500;;
  
     (** could change priority based on learned information; here I'm hard coding some changes for examples I want to work **)
let adapted_process_prop_1 p m =
  if !magic_repl_sep_inst_suggested then (*** now we believe we're working on the replacement implies separation example, so prefer propositions that take part in the proof ***)
    begin
      match !processed_neg_sep_for_pred with
      | None -> p
      | Some(a,mem,pred,bdset) -> adapted_process_prop_1_repl_imp_sep a mem pred bdset p m
    end
  else
    begin
      match !injcantor with
      | Some(a) -> adapted_process_prop_1_injcantor a p m
      | None ->
	  match (!induction,!neg_com) with
	  | (Some(nattp,zer,suc),Some(add,c)) ->
	      adapted_process_prop_1_add_com nattp zer suc add c p m
	  | _ ->
	      match m with
	      | Ap(Ap(Imp,Ap(Ap(Eq(_),x),y)),q) when x = y -> -1 (*** prefer processing formulas of the form s = s -> q : this makes the neqexample even easier ***)
	      | _ -> p
    end;;

let adapted_newinst p c m =
  match !manyeigens2 with
  | Some(a,b,g,o,r) -> (*** doing the manyeigens2 example, if c = a, then only allow g and o(g,g) as good instantiations, if c = b then only allow the good eigenvars as good insts ***)
      if c = Base(a) then
	begin
	  match m with
	  | Name(g2,_) when g2 = g -> -1
	  | Ap(Ap(Name(o2,_),Name(g2,_)),Name(g3,_)) when o2 = o && g2 = g && g3 = g -> -1
	  | _ -> 500
	end
      else if c = Base(b) then
	begin
	  match m with
	  | Name(w,_) when manyeigens2_eigen_1_p w || manyeigens2_eigen_2_p w || manyeigens2_eigen_3_p w || manyeigens2_eigen_4_p w || manyeigens2_eigen_5_p w -> -1
	  | _ -> 500
	end
      else
	p
  | None ->
      if !magic_repl_sep_inst_suggested then (*** judging instantiations for replimpsep example ***)
	begin
	  match !processed_neg_sep_for_pred with
	  | None -> p
	  | Some(settp,mem,predname,bdset) ->
	      if c = Base(settp) then (*** just filter the instantiations of base type ***)
		begin
		  match m with
		  | Name(x,_) ->
		      if Hashtbl.mem eigenvarforh x then
			-1 (*** say eigenvars are good insts (could refine this) ***)
		      else
			500 (*** all other instantiations are bad ***)
		  | _ -> 500 (*** all other instantiations are bad ***)
		end
	      else
		p (*** for other types, pass the priority without change ***)
	end
      else
	begin
	  match (!induction,!neg_com) with
	  | (Some(nattp,zer,suc),Some(add,c1)) ->
	      adapted_newinst_add_com nattp zer suc add c1 p c m
	  | _ -> p
	end
      
let rec head m =
  match m with
  | Ap(m,_) -> head m
  | _ -> m;;

let adapted_confront p a m1 m2 n1 n2 =
  match (!induction,!neg_com) with
  | (Some(nattp,zer,suc),Some(add,c)) ->
      adapted_confront_add_com nattp zer suc add c p a m1 m2 n1 n2
  | _ -> p;;
	
     (** could change priority based on learned information; here I'm hard coding some changes for examples I want to work **)
let adapted_mating p l1 l2 ml nl =
  if !magic_repl_sep_inst_suggested then (*** now we believe we're working on the replacement implies separation example, so prefer propositions that take part in the proof ***)
    begin
      match !processed_neg_sep_for_pred with
      | None -> p
      | Some(a,mem,pred,bdset) ->
	  try
	    begin
	      match head (if l1 > 0 then Hashtbl.find atomh l1 else Hashtbl.find atomh l2) with (*** I don't remember which one is positive, so ask on the fly. ***)
	      | Name(pred1,_) when pred = pred1 -> -1
	      | _ -> 500
	    end
	  with Not_found -> p
    end
  else
    p

let reset_advisor_state() =
  Hashtbl.clear atomh;
  processed_repl := None;
  processed_neg_sep_for_pred := None;
  magic_repl_sep_inst_suggested := false;
  injcantor := None;
  induction := None;
  neg_com := None;
  suggstack := []

let handlemsg msgbyte s2in s2out =
  if msgbyte = 0 then (*** General request for a suggestion for a new proposition (internal lemma/cut) or new instantiation ***)
    begin
      match !suggstack with
      | [] -> output_byte s2out 0; flush s2out (* no suggestion *)
      | SuggCut(m)::ssr -> (* send suggested cut *)
	  output_byte s2out 1;
	  trm_binout s2out m;
	  flush s2out;
	  suggstack := ssr
      | SuggInst(m)::ssr -> (* send suggested instantiation term *)
	  output_byte s2out 2;
	  trm_binout s2out m;
	  flush s2out;
	  suggstack := ssr
    end
  else if msgbyte = 4 then (*** ProcessProp1 ***)
    begin
      let p = varint_binin s2in in
      let m = trm_binin s2in in (*** could extract features of m and use that to learn something upon success ***)
      if !verbose > 9 then (Printf.printf "ProcessProp1 %d %s\n" p (trm_str m); flush stdout);
      recog_neq_example m;
      recog_manyeigens2_example m;
      recog_repl_ish m;
      recog_neg_sep_for_pred_ish m;
      suggest_repl_sep_inst();
      recog_injcantor m;
      recog_nat_ind m;
      recog_neg_com m;
      let ap = adapted_process_prop_1 p m in (** could change priority based on learned information; here I'm hard coding some changes for examples I want to work **)
      if !verbose > 9 then (Printf.printf "ProcessProp1 %s %d modified to %d\n" (trm_str m) p ap; flush stdout);
      output_byte s2out msgbyte;
      varint_binout s2out ap;
      flush s2out
    end
  else if msgbyte = 5 then (*** ProcessProp2 ***)
    begin
      let p = varint_binin s2in in
      let m = trm_binin s2in in (*** could extract features of m and use that to learn something upon success ***)
      if !verbose > 9 then (Printf.printf "ProcessProp2 %d %s\n" p (trm_str m); flush stdout);
      (** could change priority based on learned information **)
      output_byte s2out msgbyte;
      varint_binout s2out p;
      flush s2out
    end
  else if msgbyte = 6 then (*** Mating ***)
    begin
      let p = varint_binin s2in in
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      let n = varint_binin s2in in
      let ml = ref [] in
      let nl = ref [] in
      for i = 1 to n do
	ml := trm_binin s2in::!ml
      done;
      for i = 1 to n do
	nl := trm_binin s2in::!nl
      done;
      ml := List.rev !ml;
      nl := List.rev !nl;
      let ap = adapted_mating p l1 l2 ml nl in (** could change priority based on learned information; here I'm hard coding some changes for examples I want to work **)
      output_byte s2out msgbyte;
      varint_binout s2out ap;
      flush s2out
    end
  else if msgbyte = 7 then (*** Confront ***)
    begin
      let p = varint_binin s2in in
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      let a = stp_binin s2in in
      let m1 = trm_binin s2in in
      let m2 = trm_binin s2in in
      let n1 = trm_binin s2in in
      let n2 = trm_binin s2in in
      (** could change priority based on learned information **)
      output_byte s2out msgbyte;
      let ap = adapted_confront p a m1 m2 n1 n2 in
(*      if !verbose > 9 then (Printf.printf "Confront %d %d:\n%s\n%s\n%s\n%s\n" l1 l2 (trm_str m1) (trm_str n1) (trm_str m2) (trm_str n1); flush stdout); *)
      if !verbose > 9 then (Printf.printf "Confront %d %d priority %d adapted to %d\n" l1 l2 p ap);
      varint_binout s2out ap;
      flush s2out
    end
  else if msgbyte = 8 then (*** NewInst ***)
    begin
      let p = varint_binin s2in in
      let a = stp_binin s2in in
      let m = trm_binin s2in in
      (** could change priority based on learned information **)
      output_byte s2out msgbyte;
      let ap = adapted_newinst p a m in
      if !verbose > 9 then (Printf.printf "NewInst %s %s %d modified to %d\n" (stp_str a) (trm_str m) p ap; flush stdout);
      varint_binout s2out ap;
      flush s2out
    end
  else if msgbyte = 11 then (*** New Eigen ***)
    begin
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      let a = stp_binin s2in in
      let w = string_binin s2in in
      let m = Hashtbl.find atomh l1 in
      Hashtbl.add eigenvarforh w m;
      if !verbose > 9 then (Printf.printf "New eigen %s: %s\nm: %s\n" w (trm_str m) (trm_ocaml_match m); flush stdout)
    end
  else if msgbyte = 12 then (*** New Atom ***)
    begin
      let a = varint_binin s2in in
      if !verbose > 9 then (Printf.printf "Atom %d\n" a; flush stdout);
      let m = trm_binin s2in in
      if !verbose > 9 then (Printf.printf "Atom %d %s\n" a (trm_str m); flush stdout);
      Hashtbl.add atomh a m (* no reply expected here, just giving the advisor the information it needs to give advice later *)
    end
  else if msgbyte = 13 then (*** New Assumption (Unit Clause of Literal l) ***)
    begin
      let l = varint_binin s2in in
      if !verbose > 9 then (Printf.printf "Assumption %d\n" l; flush stdout);
      () (* no reply expected here, just giving the advisor the information it needs to give advice later *)
    end
  else if msgbyte = 14 then (*** New Clause ***)
    begin
      let n = varint_binin s2in in
      let c = ref [] in
      for i = 1 to n do
	c := varint_binin s2in::!c
      done;
      c := List.rev !c;
      if !verbose > 9 then (Printf.printf "Clause:"; List.iter (fun l -> Printf.printf " %d" l) !c; Printf.printf "\n"; flush stdout);
      (* no reply expected here, just giving the advisor the information it needs to give advice later *)
    end
  else if msgbyte = 15 then (*** New Inst ***)
    begin
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      let a = stp_binin s2in in
      let m = trm_binin s2in in
      if !verbose > 9 then (Printf.printf "Inst %d %d %s: %s\n" l1 l2 (stp_str a) (trm_str m); flush stdout);
      (* no reply expected here, just giving the advisor the information it needs to give advice later *)
    end
  else
    raise (Failure "unexpected msg, kill connection");;

let main port =
  let s1 = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  let ia = Unix.inet_addr_of_string "127.0.0.1" in
  Unix.bind s1 (Unix.ADDR_INET(ia, port));
  Unix.listen s1 5;
  while true do
    try
      let (s2,a2) = Unix.accept s1 in
      try
	let s2in = Unix.in_channel_of_descr s2 in
	let s2out = Unix.out_channel_of_descr s2 in
	set_binary_mode_in s2in true;
	set_binary_mode_out s2out true;
	reset_advisor_state();
	while true do
	  let msgbyte = input_byte s2in in
	  if !verbose > 5 then (Printf.printf "msg %d\n" msgbyte; flush stdout);
	  handlemsg msgbyte s2in s2out
	done
      with e ->
	if !verbose > 3 then (Printf.printf "exception %s; closing connection\n" (Printexc.to_string e); flush stdout);
	Unix.close s2
    with _ -> ()
  done;;

if Array.length Sys.argv > 1 then main (int_of_string Sys.argv.(1)) else main 2332;;
    
    

