open Syntax;;

let verbose = ref 0;;

let atomh : (int,trm) Hashtbl.t = Hashtbl.create 100;;
let eigenvarforh : (string,trm) Hashtbl.t = Hashtbl.create 100;;

let reset_learner_state () =
  Hashtbl.clear atomh;;

let handlemsg msgbyte s2in =
  if msgbyte = 4 then (*** ProcessProp1 ***)
    begin
      let p = varint_binin s2in in
      let m = trm_binin s2in in (*** could extract features of m and use that to learn something upon success ***)
      if !verbose > 9 then (Printf.printf "ProcessProp1 %d %s\n%s\n" p (trm_str m) (trm_ocaml_match m); flush stdout);
      ()
    end
  else if msgbyte = 5 then (*** ProcessProp2 ***)
    begin
      let p = varint_binin s2in in
      let m = trm_binin s2in in (*** could extract features of m and use that to learn something upon success ***)
      if !verbose > 9 then (Printf.printf "ProcessProp2 %d %s\n" p (trm_str m); flush stdout);
      ()
    end
  else if msgbyte = 6 then (*** Mating ***)
    begin
      let p = varint_binin s2in in
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      let n = varint_binin s2in in
      if !verbose > 9 then (Printf.printf "Mating %d %d\n%s\n" l1 l2 (trm_ocaml_match (Hashtbl.find atomh l1)); flush stdout);
      let ml = ref [] in
      let nl = ref [] in
      for i = 1 to n do
	ml := trm_binin s2in::!ml
      done;
      for i = 1 to n do
	nl := trm_binin s2in::!nl
      done;
      ml := List.rev !ml;
      nl := List.rev !nl
    end
  else if msgbyte = 7 then (*** Confront ***)
    begin
      let p = varint_binin s2in in
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      if !verbose > 9 then (Printf.printf "Confront %d %d\n%s\n" l1 l2 (trm_ocaml_match (Hashtbl.find atomh l1)); flush stdout);
      let a = stp_binin s2in in
      let m1 = trm_binin s2in in
      let m2 = trm_binin s2in in
      let n1 = trm_binin s2in in
      let n2 = trm_binin s2in in
      ()
    end
  else if msgbyte = 8 then (*** NewInst ***)
    begin
      let p = varint_binin s2in in
      let a = stp_binin s2in in
      let m = trm_binin s2in in
      if !verbose > 9 then (Printf.printf "NewInst %s\n%s\n" (trm_str m) (trm_ocaml_match m); flush stdout);
      ()
    end
  else if msgbyte = 11 then (*** New Eigen ***)
    begin
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      let a = stp_binin s2in in
      let w = string_binin s2in in
      if !verbose > 9 then (Printf.printf "Eigen %d %d %s: %s\n" l1 l2 (stp_str a) w; flush stdout);
      let m = Hashtbl.find atomh l1 in
      if !verbose > 9 then (Printf.printf "(EIGEN %d %d \"%s\" \"%s\")\n" l1 l2 w (String.escaped (trm_ocaml_match m)); flush stdout);
      Hashtbl.add eigenvarforh w m
    end
  else if msgbyte = 12 then (*** New Atom ***)
    begin
      let a = varint_binin s2in in
      let m = trm_binin s2in in
      if !verbose > 9 then (Printf.printf "Atom %d %s\n(ATOM %d \"%s\")\n" a (trm_str m) a (String.escaped (trm_ocaml_match m)); flush stdout);
      Hashtbl.add atomh a m
    end
  else if msgbyte = 13 then (*** New Assumption (Unit Clause of Literal l) ***)
    begin
      let l = varint_binin s2in in
      if !verbose > 9 then (Printf.printf "Assumption %d\n" l; flush stdout);
      ()
    end
  else if msgbyte = 14 then (*** New Clause ***)
    begin
      let n = varint_binin s2in in
      let c = ref [] in
      for i = 1 to n do
	c := varint_binin s2in::!c
      done;
      c := List.rev !c;
      if !verbose > 9 then (Printf.printf "Clause:"; List.iter (fun l -> Printf.printf " %d" l) !c; Printf.printf "\n"; flush stdout);
    end
  else if msgbyte = 15 then (*** New Inst ***)
    begin
      let l1 = varint_binin s2in in
      let l2 = varint_binin s2in in
      let a = stp_binin s2in in
      let m = trm_binin s2in in
      if !verbose > 9 then (Printf.printf "Inst %d %d %s: %s\n" l1 l2 (stp_str a) (trm_str m); flush stdout);
    end
  else if msgbyte = 16 then (*** minisat claims unsat; could analyze proof here to learn something ***)
    ()
  else if msgbyte = 17 then (*** E claims unsat; could try to get more information and learn something ***)
    ()
  else
    raise (Failure "unexpected msg, kill connection");;

let main port =
  let s1 = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  let ia = Unix.inet_addr_of_string "127.0.0.1" in
  Unix.bind s1 (Unix.ADDR_INET(ia, port));
  Unix.listen s1 5;
  while true do
    try
      let (s2,a2) = Unix.accept s1 in
      try
	let s2in = Unix.in_channel_of_descr s2 in
	let s2out = Unix.out_channel_of_descr s2 in
	set_binary_mode_in s2in true;
	set_binary_mode_out s2out true;
	while true do
	  let msgbyte = input_byte s2in in
	  if !verbose > 5 then (Printf.printf "msg %d\n" msgbyte; flush stdout);
	  handlemsg msgbyte s2in
	done
      with e ->
	if !verbose > 3 then (Printf.printf "exception %s; closing connection\n" (Printexc.to_string e); flush stdout);
	Unix.close s2
    with _ -> ()
  done;;

if Array.length Sys.argv > 1 then main (int_of_string Sys.argv.(1)) else main 2323;;
    
    

